import glob
import logging
import os.path
import re

import numpy as np
import pandas as pd
import pytest
from click.testing import CliRunner
from pymagicc.io import MAGICCData

from netcdf_scm.cli import stitch_netcdf_scm_ncs
from netcdf_scm.io import load_scmdataframe


def _do_generic_stitched_data_tests(stiched_scmdf):
    assert stiched_scmdf["scenario"].nunique() == 1
    assert "(child) branch_time_in_parent" in stiched_scmdf.metadata
    assert "(child) parent_experiment_id" in stiched_scmdf.metadata
    assert "(parent) source_id" in stiched_scmdf.metadata
    assert "(parent) experiment_id" in stiched_scmdf.metadata


def _do_generic_normalised_data_tests(normalised_scmdf):
    assert normalised_scmdf["scenario"].nunique() == 1
    assert "(child) branch_time_in_parent" in normalised_scmdf.metadata
    assert "(child) parent_experiment_id" in normalised_scmdf.metadata
    assert "(normalisation) source_id" in normalised_scmdf.metadata
    assert "(normalisation) experiment_id" in normalised_scmdf.metadata
    assert "normalisation method" in normalised_scmdf.metadata


def _check_expected_values(res, other, years, norm_shift=None, regions=None):
    if regions is None:
        regions = ["World", "World|North Atlantic Ocean", "World|Southern Hemisphere"]

    for region in regions:
        res_vals = res.filter(region=region, year=years).values
        assert not any([s == 0 for s in res_vals.shape])

        other_vals = other.filter(region=region, year=years).values
        assert not any([s == 0 for s in other_vals.shape])

        if norm_shift is not None:
            other_vals -= norm_shift[
                norm_shift.index.get_level_values("region") == region
            ].values.squeeze()

        np.testing.assert_allclose(
            res_vals, other_vals, rtol=1e-5, atol=1e-6,
        )


def _do_filepath_time_consistency_check(res, filepath, annual=False):
    if annual:
        expected_end = "{:04d}-{:04d}.MAG".format(
            res["time"].min().year, res["time"].max().year,
        )
    else:
        expected_end = "{:04d}{:02d}-{:04d}{:02d}.MAG".format(
            res["time"].min().year,
            res["time"].min().month,
            res["time"].max().year,
            res["time"].max().month,
        )
    assert filepath.endswith(expected_end)


def test_stitching_default(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_default"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*EC-Earth3-Veg.*ssp585.*r1i1p1f1.*hfds.*",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])
    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp585" in child_path
    assert "r1i1p1f1" in child_path
    assert "EC-Earth3-Veg" in child_path
    assert "hfds" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(res, child, range(2015, 2017))
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2015, 2017):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in parent_path
    assert "EC-Earth3-Veg" in parent_path
    assert "hfds" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(res, parent, range(2013, 2015))
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2013, 2015):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             parent.filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )


def test_stitching_two_levels(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_two_levels"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                "^(?!.*(NCAR)).*GeoMIP.*G6solar.*$",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "G6solar" in child_path
    assert "r1i1p1f2" in child_path
    assert "UKESM1-0-LL" in child_path
    assert "v20191031" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(
        res,
        child,
        range(2020, 2100),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2020, 2100):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "ssp585" in parent_path
    assert "r1i1p1f2" in parent_path
    assert "UKESM1-0-LL" in parent_path
    assert "v20190726" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(
        res,
        parent,
        range(2015, 2020),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2015, 2020):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             parent.filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )

    grandparent_path = res.metadata["(grandparent) netcdf-scm crunched file"]
    assert "historical" in grandparent_path
    assert "r1i1p1f2" in grandparent_path
    assert "UKESM1-0-LL" in grandparent_path
    assert "v20190627" in grandparent_path

    grandparent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", grandparent_path))[0]
    )

    _check_expected_values(
        res,
        grandparent,
        range(1880, 2014),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(1880, 2014):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             grandparent.filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )


def test_stitching_in_file_BCC_CSM2_MR(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_in_file_BCC_CSM2_MR"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*r1i1p1f1.*tas.*",
                "--out-format",
                "mag-files-average-year-mid-year",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0], annual=True)

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp126" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(res, child.time_mean("AC"), range(2015, 2017))
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2015, 2017):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.time_mean("AC").filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in parent_path
    assert "BCC-CSM2-MR" in parent_path
    assert "tas" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(res, parent.time_mean("AC"), range(2013, 2015))
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2013, 2015):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             parent.time_mean("AC").filter(region=region, year=y).values,
    #             rtol=1e-5,
    #         )

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    assert len(bcc_warning) == 1
    assert bcc_warning[0][1] == logging.WARNING


def test_stitching_no_parent(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_no_parent"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CNRM-ESM2-1.*r2i1p1f2.*/cSoil/.*",
            ],
        )

    assert result.exit_code != 0

    error_msg = re.compile(
        ".*No parent data \\(ssp585\\) available for "
        ".*CMIP6/ScenarioMIP/CNRM-CERFACS/CNRM-ESM2-1/ssp534-over/r2i1p1f2/Emon/cSoil/gr/v20190410/netcdf-scm_cSoil_Emon_CNRM-ESM2-1_ssp534-over_r2i1p1f2_gr_201501-210012.nc"
        ", we looked in "
        ".*CMIP6/ScenarioMIP/CNRM-CERFACS/CNRM-ESM2-1/ssp585/r2i1p1f2/Emon/cSoil/gr/\\*/netcdf-scm_cSoil_Emon_CNRM-ESM2-1_ssp585_r2i1p1f2_gr_\\*.nc"
    )
    no_parent_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_parent_error) == 1
    assert no_parent_error[0][1] == logging.ERROR


def test_stitching_historical_only(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_historical_only"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*GFDL-CM4.*1pctCO2.*r1i1p1f1.*",
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    # should just be same as wrangle if like this
    assert not any(["(child)" in k for k in res.metadata])
    error_msg = re.compile(
        ".*No normalisation is being done and the parent of 1pctCO2 is piControl for "
        "infile: "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/1pctCO2/r1i1p1f1/Amon/tas/gr1/v20180701/netcdf-scm_tas_Amon_GFDL-CM4_1pctCO2_r1i1p1f1_gr1_000101-015012.nc"
    )
    no_operation_info = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_operation_info) == 1
    assert no_operation_info[0][1] == logging.INFO


def test_stitching_esm_hist_only(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_esm_hist_only"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CanESM5.*esm-hist.*r1i1p1f1.*fgco2.*",
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    # should just be same as wrangle if like this
    assert not any(["(child)" in k for k in res.metadata])
    error_msg = re.compile(
        "No normalisation is being done and the parent of esm-hist is esm-piControl "
        "for infile: "
        ".*CMIP6/CMIP/CCCma/CanESM5/esm-hist/r1i1p1f1/Omon/fgco2/gn/v20190429/netcdf-scm_fgco2_Omon_CanESM5_esm-hist_r1i1p1f1_gn_185001-201412.nc"
    )
    no_operation_info = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_operation_info) == 1
    assert no_operation_info[0][1] == logging.INFO


def test_stitching_with_normalisation(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation"
    norm_method = "31-yr-mean-after-branch-time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CESM2.*r10i1p1f1.*tas.*",
                "--normalise",
                norm_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == norm_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "historical" in child_path
    assert "r10i1p1f1" in child_path
    assert "CESM2" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CESM2" in normalisation_path
    assert "tas" in normalisation_path

    normalisation = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", normalisation_path))[0]
    )

    assert child.metadata["branch_time_in_parent"] == 306600.0
    assert child.metadata["parent_time_units"] == "days since 0001-01-01 00:00:00"
    expected_norm_year_raw = int(1 + (306600 / 365))

    norm_shift = (
        normalisation.filter(
            year=range(expected_norm_year_raw, expected_norm_year_raw + 31)
        )
        .timeseries()
        .mean(axis=1)
    )

    _check_expected_values(res, child, range(1850, 2015), norm_shift=norm_shift)
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(1850, 2015):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.filter(region=region, year=y).values
    #             - norm_shift[
    #                 norm_shift.index.get_level_values("region") == region
    #             ].values.squeeze(),
    #             rtol=1e-5,
    #         )


def test_stitching_with_normalisation_esm_hist(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_esm_hist"
    norm_method = "31-yr-mean-after-branch-time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CanESM5.*esm-hist.*r1i1p1f1.*fgco2.*",
                "--normalise",
                norm_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == norm_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "esm-hist" in child_path
    assert "r1i1p1f1" in child_path
    assert "CanESM5" in child_path
    assert "fgco2" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "esm-piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CanESM5" in normalisation_path
    assert "fgco2" in normalisation_path

    normalisation = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", normalisation_path))[0]
    )

    assert child.metadata["branch_time_in_parent"] == 1277865.0
    assert child.metadata["parent_time_units"] == "days since 1850-01-01 0:0:0.0"
    expected_norm_year_raw = int(1850 + (1277865.0 / 365))

    norm_shift = (
        normalisation.filter(
            year=range(expected_norm_year_raw, expected_norm_year_raw + 31)
        )
        .timeseries()
        .mean(axis=1)
    )

    _check_expected_values(res, child, range(1850, 2015), norm_shift=norm_shift)
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in :
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.filter(region=region, year=y).values
    #             - norm_shift[
    #                 norm_shift.index.get_level_values("region") == region
    #             ].values.squeeze(),
    #             rtol=1e-5,
    #         )


def test_stitching_with_normalisation_year_zero(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_year_zero"
    norm_method = "31-yr-mean-after-branch-time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CESM2.*abrupt-4xCO2.*r1i1p1f1.*fgco2.*",
                "--normalise",
                norm_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == norm_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "abrupt-4xCO2" in child_path
    assert "r1i1p1f1" in child_path
    assert "CESM2" in child_path
    assert "fgco2" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CESM2" in normalisation_path
    assert "fgco2" in normalisation_path

    normalisation = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", normalisation_path))[0]
    )

    assert child.metadata["branch_time_in_parent"] == 182865.0
    assert child.metadata["parent_time_units"] == "days since 0001-01-01 00:00:00"
    expected_norm_year_raw = int(+(182865.0 / 365))

    norm_shift = (
        normalisation.filter(
            year=range(expected_norm_year_raw, expected_norm_year_raw + 31)
        )
        .timeseries()
        .mean(axis=1)
    )

    _check_expected_values(res, child, range(1, 1000), norm_shift=norm_shift)


def test_stitching_with_normalisation_two_levels(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation"
    norm_method = "31-yr-mean-after-branch-time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*tas.*",
                "--normalise",
                norm_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == norm_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp126" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "BCC-CSM2-MR" in normalisation_path
    assert "tas" in normalisation_path

    normalisation = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", normalisation_path))[0]
    )

    assert parent.metadata["branch_time_in_parent"] == 2289.0
    assert parent.metadata["parent_time_units"] == "days since 1850-01-01"
    # assuming mislabelling of years as days
    expected_norm_year_raw = 2289

    norm_shift = (
        normalisation.filter(
            year=range(expected_norm_year_raw, expected_norm_year_raw + 31)
        )
        .timeseries()
        .mean(axis=1)
    )

    _check_expected_values(res, child, range(2015, 2101), norm_shift=norm_shift)
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(2015, 2101):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.filter(region=region, year=y).values
    #             - norm_shift[
    #                 norm_shift.index.get_level_values("region") == region
    #             ].values.squeeze(),
    #             rtol=1e-5,
    #         )

    _check_expected_values(res, parent, range(1850, 2015), norm_shift=norm_shift)
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(1850, 2015):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             parent.filter(region=region, year=y).values
    #             - norm_shift[
    #                 norm_shift.index.get_level_values("region") == region
    #             ].values.squeeze(),
    #             rtol=1e-5,
    #         )

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    assert len(bcc_warning) == 2
    assert bcc_warning[0][1] == logging.WARNING


def test_stitching_with_normalisation_in_file_BCC_CSM2_MR(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_in_file_BCC_CSM2_MR"
    norm_method = "31-yr-mean-after-branch-time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*1pctCO2-bgc.*r1i1p1f1.*tas.*",
                "--out-format",
                "mag-files-average-year-mid-year",
                "--normalise",
                norm_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    assert res.metadata["normalisation method"] == norm_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "1pctCO2-bgc" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "BCC-CSM2-MR" in normalisation_path
    assert "tas" in normalisation_path

    normalisation = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", normalisation_path))[0]
    )

    assert child.metadata["branch_time_in_parent"] == 0.0
    assert child.metadata["parent_time_units"] == "days since 1850-01-01"
    expected_norm_year_raw = 1850

    norm_shift = (
        normalisation.filter(
            year=range(expected_norm_year_raw, expected_norm_year_raw + 31)
        )
        .timeseries()
        .mean(axis=1)
    )

    _check_expected_values(
        res, child.time_mean("AC"), range(1850, 2015), norm_shift=norm_shift
    )
    # for region in ["World", "World|North Atlantic Ocean"]:
    #     for y in range(1850, 2015):
    #         np.testing.assert_allclose(
    #             res.filter(region=region, year=y).values,
    #             child.time_mean("AC").filter(region=region, year=y).values
    #             - norm_shift[
    #                 norm_shift.index.get_level_values("region") == region
    #             ].values.squeeze(),
    #             rtol=1e-5,
    #         )

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    # this file has a branch time of zero so the warning shouldn't be zero
    assert not bcc_warning


def test_stitching_with_normalisation_no_picontrol(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_no_picontrol"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*GFDL-CM4.*1pctCO2.*r1i1p1f1.*",
                "--normalise",
                "31-yr-mean-after-branch-time",
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*No parent data \\(piControl\\) available for "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/1pctCO2/r1i1p1f1/Amon/tas/gr1/v20180701/netcdf-scm_tas_Amon_GFDL-CM4_1pctCO2_r1i1p1f1_gr1_000101-015012.nc"
        ", we looked in "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/piControl/r1i1p1f1/Amon/tas/gr1/\\*/netcdf-scm_tas_Amon_GFDL-CM4_piControl_r1i1p1f1_gr1_\\*.nc"
    )
    no_parent_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_parent_error) == 1
    assert no_parent_error[0][1] == logging.ERROR


def test_stitching_with_normalisation_no_branching_time(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_no_branching_time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CNRM-CM6-1.*hist-aer.*tas.*",
                "--normalise",
                "31-yr-mean-after-branch-time",
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*Branching time `188301` not available in piControl data in "
        "CMIP6/CMIP/CNRM-CERFACS/CNRM-CM6-1/piControl/r1i1p1f2/Amon/tas/gr/v20180814/netcdf-scm_tas_Amon_CNRM-CM6-1_piControl_r1i1p1f2_gr_230001-231012.nc"
    )
    no_branch_time_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_branch_time_error) == 1
    assert no_branch_time_error[0][1] == logging.ERROR


def test_stitching_with_normalisation_not_enough_branching_time(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_no_branching_time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*MIROC6.*r1i1p1f1.*rlut.*",
                "--normalise",
                "31-yr-mean-after-branch-time",
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*Only `320001` to `320212` is available after the branching time `320001` in piControl "
        "data in "
        "CMIP6/CMIP/MIROC/MIROC6/piControl/r1i1p1f1/Amon/rlut/gn/v20181212/netcdf-scm_rlut_Amon_MIROC6_piControl_r1i1p1f1_gn_320001-320212.nc"
    )
    not_enough_norm_data_error = [
        r for r in caplog.record_tuples if error_msg.match(r[2])
    ]
    assert len(not_enough_norm_data_error) == 1
    assert not_enough_norm_data_error[0][1] == logging.ERROR


@pytest.mark.parametrize(
    "out_format",
    (
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_stitching_file_types(tmpdir, caplog, test_cmip6_crunch_output, out_format):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "--out-format",
                out_format,
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*(ssp126|historical).*tas.*",
                "--normalise",
                "31-yr-mean-after-branch-time",
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(
        os.path.join(
            output_dir, "flat", "*.IN" if out_format.startswith("magicc") else "*.MAG"
        )
    )
    assert len(out_files) == 4 if out_format.startswith("magicc") else 2

    for p in out_files:
        res = MAGICCData(p)
        _do_generic_normalised_data_tests(res)
        if "ssp26" in p:
            _do_generic_stitched_data_tests(res)
        if p.endswith(".MAG"):
            _do_filepath_time_consistency_check(res, p, annual="year" in out_format)


@pytest.mark.parametrize(
    "out_format",
    (
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_prefix(tmpdir, caplog, test_cmip6_crunch_output, out_format):
    output_dir = str(tmpdir)
    crunch_contact = "test_prefix"
    prefix = "NORMED"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*tas.*",
                "--out-format",
                out_format,
                "--prefix",
                prefix,
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(
        os.path.join(
            output_dir, "flat", "*.IN" if out_format.startswith("magicc") else "*.MAG"
        )
    )
    assert len(out_files) == 2 if out_format.startswith("magicc") else 1
    assert [os.path.basename(p).startswith("{}_".format(prefix)) for p in out_files]


def test_target_units(tmpdir, caplog, test_cmip6_crunch_output):
    target_unit = "J / yr / m^2"
    target_units = pd.DataFrame(
        [["fgco2", "g / m**2 / s"], ["hfds", target_unit]], columns=["variable", "unit"]
    )
    target_units_csv = os.path.join(tmpdir, "target_units.csv")
    target_units.to_csv(target_units_csv, index=False)

    output_dir = str(tmpdir)
    crunch_contact = "test_target_units"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*EC-Earth3-Veg.*ssp585.*r1i1p1f1.*hfds.*",
                "--target-units-specs",
                target_units_csv,
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "flat", "*.MAG"))
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res["unit"].unique().tolist() == ["Jperyrpermsuper2"]
    assert "Converting units" in result.stderr



import pytest


@pytest.mark.script_launch_mode("subprocess")
@pytest.mark.parametrize(
    "entry_point", ("netcdf-scm-crunch", "netcdf-scm-wrangle", "netcdf-scm-stitch",)
)
def test_silicone_explore_quantiles_rolling_windows(entry_point, script_runner):
    res = script_runner.run(entry_point, "--help")
    assert res.success
    assert res.stdout
    assert res.stderr == ""



"""
Wrappers of the iris cube.

These classes automate handling of a number of netCDF processing steps.
For example, finding surface land fraction files, applying regions to data and
returning timeseries in key regions for simple climate models.
"""
import logging
import os
import re
import warnings
from abc import ABC, abstractmethod, abstractproperty
from datetime import datetime
from os.path import basename, dirname, join, splitext

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from scmdata import ScmDataFrame

from . import __version__
from .definitions import _LAND_FRACTION_REGIONS, _SCM_TIMESERIES_META_COLUMNS
from .utils import (
    _check_cube_and_adjust_if_needed,
    _vector_cftime_conversion,
    assert_all_time_axes_same,
    cube_lat_lon_grid_compatible_with_array,
    get_cube_timeseries_data,
    get_scm_cube_time_axis_in_calendar,
    take_lat_lon_mean,
    unify_lat_lon,
)
from .weights import DEFAULT_REGIONS, CubeWeightCalculator

try:
    import cftime

    import iris
    import iris.analysis.cartography
    import iris.coord_categorisation
    import iris.experimental.equalise_cubes
    from iris.exceptions import CoordinateMultiDimError, ConcatenateError
    from iris.fileformats import netcdf
    from iris.util import unify_time_units

    # monkey patch netCDF4 loading to avoid very small chunks
    # required until there is a resolution to
    # https://github.com/SciTools/iris/issues/3333
    # and
    # https://github.com/SciTools/iris/issues/3357
    def _get_cf_var_data(cf_var, filename):
        import netCDF4  # pylint:disable=import-outside-toplevel

        # Get lazy chunked data out of a cf variable.
        dtype = netcdf._get_actual_dtype(cf_var)  # pylint:disable=protected-access

        # Create cube with deferred data, but no metadata
        fill_value = getattr(
            cf_var.cf_data,
            "_FillValue",
            netCDF4.default_fillvals[cf_var.dtype.str[1:]],  # pylint:disable=no-member
        )
        proxy = netcdf.NetCDFDataProxy(
            cf_var.shape, dtype, filename, cf_var.cf_name, fill_value
        )
        return netcdf.as_lazy_data(proxy, chunks=None)

    netcdf._get_cf_var_data = _get_cf_var_data  # pylint:disable=protected-access

except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()

logger = logging.getLogger(__name__)


class SCMCube:  # pylint:disable=too-many-public-methods
    """
    Class for processing netCDF files for use in simple climate models.

    Common, shared operations are implemented here.
    However, methods like ``_get_data_directory`` raise ``NotImplementedError``
    because these are always context dependent.
    Hence to use this base class, you must use a subclass of it which defines these
    context specific methods.
    """

    cube = None
    """
    :obj:`iris.cube.Cube`: The Iris cube which is wrapped by this :obj:`SCMCube` instance.
    """

    lat_name = "latitude"
    """str: The expected name of the latitude co-ordinate in data."""

    lon_name = "longitude"
    """str: The expected name of the longitude co-ordinate in data."""

    time_name = "time"
    """str: The expected name of the time co-ordinate in data."""

    time_period_separator = "-"
    """
    str: Character used to separate time period strings in the time period indicator in filenames.

    e.g. ``-`` is the 'time period separator' in "2015-2030".
    """

    _time_period_regex = None

    _known_timestamps = {
        4: {"datetime_str": "%Y", "expected_timestep": relativedelta(years=+1)},
        6: {"datetime_str": "%Y%m", "expected_timestep": relativedelta(months=+1)},
        8: {"datetime_str": "%Y%m%d", "expected_timestep": relativedelta(days=+1)},
        10: {"datetime_str": "%Y%m%d%H", "expected_timestep": relativedelta(hours=+1)},
    }
    _timestamp_definitions = None

    _scm_timeseries_id_map = {
        "climate_model": "model",
        "scenario": "experiment",
        "activity_id": "activity",
        "member_id": "ensemble_member",
        "mip_era": "mip_era",
        "variable": "variable_name",
    }
    """Mapping from cube attributes (derived from read files) to SCM timeseries metadata"""

    _weight_calculator = None
    """:obj:`CubeWeightCalculator` to use to mask self"""

    _realm_key = "realm"
    """str: key which identifies the realm of the data"""

    _have_guessed_realm = False
    """bool: Have we already guessed our cube data's realm?"""

    _area_weights_units = "m**2"
    """str: assumed units for area weights (checked when data is read)"""

    def __init__(self):
        self._loaded_paths = []
        self._metadata_cubes = {}
        self._weight_calculator = None

    @property
    def netcdf_scm_realm(self):
        """
        str: The realm in which NetCDF-SCM thinks the data belongs.

        This is used to make decisions about how to take averages of the data and
        where to find metadata variables.

        If it is not sure, NetCDF-SCM will guess that the data belongs to the
        'atmosphere' realm.
        """
        try:
            if self.cube.attributes[self._realm_key] in ("ocean", "ocnBgchem"):
                return "ocean"
            if self.cube.attributes[self._realm_key] in ("land"):
                return "land"
            if self.cube.attributes[self._realm_key] in ("atmos"):
                return "atmosphere"
            if not self._have_guessed_realm:
                logger.info(
                    "Unrecognised `%s` attribute value, `%s`, in `self.cube`, NetCDF-SCM will treat the data "
                    "as `atmosphere`",
                    self._realm_key,
                    self.cube.attributes[self._realm_key],
                )
                self._have_guessed_realm = True
            return "atmosphere"
        except KeyError:
            if not self._have_guessed_realm:
                logger.info(
                    "No `%s` attribute in `self.cube`, NetCDF-SCM will treat the data "
                    "as `atmosphere`",
                    self._realm_key,
                )
                self._have_guessed_realm = True
            return "atmosphere"

    @property
    def areacell_var(self):
        """
        str: The name of the variable associated with the area of each gridbox.

        If required, this is used to determine the area of each cell in a data file. For
        example, if our data file is ``tas_Amon_HadCM3_rcp45_r1i1p1_200601.nc`` then
        ``areacell_var`` can be used to work  out the name of the associated cell area
        file. In some cases, it might be as simple as replacing ``tas`` with the value of
        ``areacell_var``.
        """
        if self.netcdf_scm_realm in ("ocean",):
            return "areacello"

        return "areacella"

    @property
    def surface_fraction_var(self):
        """
        str: The name of the variable associated with the surface fraction in each gridbox.

        If required, this is used when looking for the surface fraction file which
        belongs to a given data file. For example, if our data file is
        ``tas_Amon_HadCM3_rcp45_r1i1p1_200601.nc`` then ``surface_fraction_var`` can
        be used to work out the name of the associated surface fraction file. In some
        cases, it might be as simple as replacing ``tas`` with the value of
        ``surface_fraction_var``.
        """
        if self.netcdf_scm_realm in ("ocean",):
            return "sftof"

        return "sftlf"

    @property
    def table_name_for_metadata_vars(self):
        """
        str: The name of the 'table' in which metadata variables can be found.

        For example, ``fx`` or ``Ofx``.

        We wrap this as a property as table typically means ``table_id`` but is
        sometimes referred to in other ways e.g. as ``mip_table`` in CMIP5.
        """
        return self._table_name_for_metadata_vars

    @property
    def _table_name_for_metadata_vars(self):
        if self.netcdf_scm_realm in ("ocean",):
            return "Ofx"

        return "fx"

    @property
    def time_period_regex(self):
        """
        :obj:`_sre.SRE_Pattern`: Regular expression which captures the timeseries identifier in input data files.

        For help on regular expressions, see :ref:`regular expressions <regular-expressions>`.
        """
        if self._time_period_regex is None:
            self._time_period_regex = re.compile(
                r".*_((\d*)" + re.escape(self.time_period_separator) + r"?(\d*)?).*"
            )
        return self._time_period_regex

    @property
    def timestamp_definitions(self):
        """
        dict: Definition of valid timestamp information and corresponding key values.

        This follows the CMIP standards where time strings must be one of the
        following: YYYY, YYYYMM, YYYYMMDD, YYYYMMDDHH or one of the previous combined
        with a hyphen e.g. YYYY-YYYY.

        Each key in the definitions dictionary is the length of the timestamp. Each
        value is itself a dictionary, with keys:

        - datetime_str: the string required to convert a timestamp of this length into a datetime using ``datetime.datetime.strptime``
        - generic_regexp: a regular expression which will match timestamps in this format
        - expected_timestep: a ``dateutil.relativedelta.relativedelta`` object which contains the expected timestep in files with this timestamp

        Examples
        --------
        >>> self.timestamp_definitions[len("2012")]["datetime_str"]
        "%Y"
        """
        if self._timestamp_definitions is None:
            self._timestamp_definitions = {}
            for key, value in self._known_timestamps.items():
                self._timestamp_definitions[key] = value
                generic_regexp = r"\d{" + "{}".format(key) + r"}"
                self._timestamp_definitions[key]["generic_regexp"] = generic_regexp
                hyphen_key = 2 * key + 1
                self._timestamp_definitions[hyphen_key] = {
                    "datetime_str": "{0}-{0}".format(value["datetime_str"]),
                    "expected_timestep": value["expected_timestep"],
                    "generic_regexp": "{0}-{0}".format(generic_regexp),
                }

        return self._timestamp_definitions

    @property
    def dim_names(self):
        """
        list: Names of the dimensions in this cube

        Here the names are the ``standard_names`` which means there can be
        ``None`` in the output.
        """
        return [c.standard_name for c in self.cube.coords()]

    @property
    def lon_dim(self):
        """:obj:`iris.coords.DimCoord` The longitude dimension of the data."""
        return self.cube.coord(self.lon_name)

    @property
    def lon_dim_number(self):
        """
        int: The index which corresponds to the longitude dimension.

        e.g. if longitude is the third dimension of the data, then
        ``self.lon_dim_number`` will be ``2`` (Python is zero-indexed).
        """
        return self.dim_names.index(self.lon_name)

    @property
    def lat_dim(self):
        """:obj:`iris.coords.DimCoord` The latitude dimension of the data."""
        return self.cube.coord(self.lat_name)

    @property
    def lat_dim_number(self):
        """
        int: The index which corresponds to the latitude dimension.

        e.g. if latitude is the first dimension of the data, then
        ``self.lat_dim_number`` will be ``0`` (Python is zero-indexed).
        """
        return self.dim_names.index(self.lat_name)

    @property
    def lat_lon_shape(self):
        """
        tuple: 2D Tuple of ``int`` which gives shape of a lat-lon slice of the data

        e.g. if the cube's shape is (4, 3, 5, 4) and its dimensions are (time, lat,
        depth, lon) then ``cube.lat_lon_shape`` will be ``(3, 4)``

        Raises
        ------
        AssertionError
            No lat lon slice can be deduced (if this happens, please raise an issue at
            https://gitlab.com/znicholls/netcdf-scm/issues so we can address your use
            case).
        """
        return self._get_cube_lat_lon_shape()

    def _get_cube_lat_lon_shape(self):
        # # we can't just do this as it doesn't work for non-orthogonal co-ordinates...
        # lat_lon_slice = next(self.cube.slices([self.lat_name, self.lon_name]))

        non_lat_lon_dims = self._get_non_lat_lon_dims()
        lat_lon_slice = next(self.cube.slices_over(non_lat_lon_dims))

        if not lat_lon_slice.shape:  # pragma: no cover # emergency valve
            raise AssertionError(
                "No lat lon slice data found, please raise an issue at "
                "https://gitlab.com/znicholls/netcdf-scm/issues"
            )

        return lat_lon_slice.shape

    def _get_non_lat_lon_dims(self):
        coords_match_shape = len(self.cube.coords()) == len(self.cube.shape)
        if coords_match_shape:
            # we can happily assume that no funny business is required
            return [
                c.standard_name if c.standard_name is not None else c.long_name
                for c in self.cube.coords()
                if c.standard_name not in [self.lat_name, self.lon_name]
            ]

        # try and make sense of what we have
        # firstly try dropping out any `cell index` co-ordinates
        cell_index_coords = [
            c.var_name
            for c in self.cube.coords()
            if c.standard_name is None and c.long_name.startswith("cell index")
        ]
        coords_stripped = [
            c for c in self.cube.coords() if c.var_name not in cell_index_coords
        ]
        coords_stripped_match_shape = len(coords_stripped) == len(self.cube.shape)
        if coords_stripped_match_shape:
            logger.debug(
                "co-ordinates make sense once we ignore `%s`", cell_index_coords,
            )
            return [
                c.standard_name if c.standard_name is not None else c.long_name
                for c in coords_stripped
                if c.standard_name not in [self.lat_name, self.lon_name]
            ]

        # try just looking at the dimensions which have more than one value i.e.
        # ignore scalar dimensions
        multi_value_dims = [c for c in coords_stripped if c.points.shape != (1,)]

        if len(multi_value_dims) == len(  # pragma: no cover # emergency valve
            self.cube.shape
        ):
            single_value_coords = [
                c.var_name for c in coords_stripped if c.points.shape == (1,)
            ]
            logger.debug(
                "co-ordinates make sense once we ignore `%s` and "
                "single value co-ordinates `%s`",
                cell_index_coords,
                single_value_coords,
            )
            return [
                c.standard_name if c.standard_name is not None else c.long_name
                for c in multi_value_dims
                if c.standard_name not in [self.lat_name, self.lon_name]
            ]

        # throw a warning as we don't know what's going on here
        logger.warning(
            "we don't know what we're looking at, `_get_non_lat_lon_dims` "
            "might not work"
        )

        # return whatever we have
        return [
            c.standard_name if c.standard_name is not None else c.long_name
            for c in self.cube.coords()
            if c.standard_name not in [self.lat_name, self.lon_name]
        ]

    @property
    def time_dim(self):
        """:obj:`iris.coords.DimCoord` The time dimension of the data."""
        return self.cube.coord(self.time_name)

    @property
    def time_dim_number(self):
        """
        int: The index which corresponds to the time dimension.

        e.g. if time is the first dimension of the data, then
        ``self.time_dim_number`` will be ``0`` (Python is zero-indexed).
        """
        return self.dim_names.index(self.time_name)

    @property
    def info(self):
        """
        dict: Information about the cubes source files

        ``res["files"]`` contains the files used to load the data in this cube.
        ``res["metadata"]`` contains information for each of the metadata cubes used to
        load the data in this cube.
        """
        r = {"files": self._loaded_paths}
        if self._metadata_cubes:
            #  Get the info dict for each of the metadata cubes
            r["metadata"] = {k: v.info for k, v in self._metadata_cubes.items()}
        return r

    def _load_cube(self, filepath, constraint=None):
        logger.debug("loading cube %s", filepath)
        self._loaded_paths.append(filepath)
        # Raises Warning and Exceptions
        self.cube = _check_cube_and_adjust_if_needed(
            iris.load_cube(filepath, constraint=constraint)
        )

    def load_data_from_path(
        self, filepath, process_warnings=True  # pylint:disable=unused-argument
    ):
        """
        Load data from a path.

        If you are using the ``SCMCube`` class directly, this method simply loads the
        path into an iris cube which can be accessed through ``self.cube``.

        If implemented on a subclass of ``SCMCube``, this method should:

        - use ``self.get_load_data_from_identifiers_args_from_filepath`` to determine the suitable set of arguments to pass to ``self.load_data_from_identifiers`` from the filepath
        - load the data using ``self.load_data_from_identifiers`` as this method contains much better checks and helper components

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.

        process_warnings : bool
            Should I process warnings to add e.g. missing metadata information?
        """
        self._load_cube(filepath)

    def load_data_in_directory(self, directory=None, process_warnings=True):
        """
        Load data in a directory.

        The data is loaded into an iris cube which can be accessed through
        ``self.cube``.

        Initially, this method is intended to only be used to load data when it is
        saved in a number of different timeslice files e.g.:

        - tas_Amon_HadCM3_rcp45_r1i1p1_200601-203012.nc
        - tas_Amon_HadCM3_rcp45_r1i1p1_203101-203512.nc
        - tas_Amon_HadCM3_rcp45_r1i1p1_203601-203812.nc

        It is not intended to be used to load multiple different variables or
        non-continuous timeseries. These use cases could be added in future, but are
        not required yet so have not been included.

        Note that this function removes any attributes which aren't common between the
        loaded cubes. In general, we have found that this mainly means
        ``creation_date``, ``tracking_id`` and ``history`` are deleted. If unsure,
        please check.

        Parameters
        ----------
        directory : str
            Directory from which to load the data.

        process_warnings : bool
            Should I process warnings to add e.g. missing metadata information?

        Raises
        ------
        ValueError
            If the files in the directory are not from the same run (i.e. their
            filenames are not identical except for the timestamp) or if the files
            don't form a continuous timeseries.
        """
        self._load_and_concatenate_files_in_directory(
            directory, process_warnings=process_warnings
        )

    def _load_and_concatenate_files_in_directory(
        self, directory, process_warnings=True
    ):
        self._check_data_names_in_same_directory(directory)

        # we use a loop here to make the most of finding missing data like
        # land-surface fraction and cellarea, something iris can't automatically do
        loaded_cubes_iris = iris.cube.CubeList()
        for f in sorted(os.listdir(directory)):
            self.load_data_from_path(
                join(directory, f), process_warnings=process_warnings
            )
            loaded_cubes_iris.append(self.cube)

        unify_time_units(loaded_cubes_iris)
        unify_lat_lon(loaded_cubes_iris)
        iris.experimental.equalise_cubes.equalise_attributes(loaded_cubes_iris)
        try:
            self.cube = loaded_cubes_iris.concatenate_cube()
        except ConcatenateError:
            long_names = []
            logger.warning(
                "Concatenating cubes in %s failed, attempting workarounds", directory
            )
            logger.warning("Removing `time_origin` attribute from all cubes")
            for ec in loaded_cubes_iris:
                long_names.append(ec.long_name)
                # try popping off the time_origin attribute
                ec.coord("time").attributes.pop("time_origin", None)

            long_name_majority_rules = max(set(long_names), key=long_names.count)
            # also try making sure that all long_name are the same
            logger.warning(
                "Forcing `long_name` of all cubes to be `%s`", long_name_majority_rules
            )
            for ec in loaded_cubes_iris:
                ec.long_name = long_name_majority_rules

            self.cube = loaded_cubes_iris.concatenate_cube()

    def _check_data_names_in_same_directory(self, directory):
        found_files = sorted(os.listdir(directory))

        assertion_error_msg = (
            "Cannot join files in:\n"
            "{}\n"
            "Files found:\n"
            "- {}".format(directory, "\n- ".join(found_files))
        )

        base_regexp = self._get_timestamp_regex_from_filename(found_files[0])
        expected_timestep = self._get_expected_timestep_from_filename(found_files[0])

        file_timestamp_bits_prev = self._get_timestamp_bits_from_filename(
            found_files[0]
        )
        time_format = self.timestamp_definitions[
            len(file_timestamp_bits_prev["timestart_str"])
        ]["datetime_str"]
        for found_file in found_files[1:]:
            if not re.match(base_regexp, found_file):
                raise AssertionError(assertion_error_msg)

            file_timestamp_bits = self._get_timestamp_bits_from_filename(found_file)
            end_time_prev = datetime.strptime(
                file_timestamp_bits_prev["timeend_str"], time_format
            )
            start_time = datetime.strptime(
                file_timestamp_bits["timestart_str"], time_format
            )

            if relativedelta(start_time, end_time_prev) != expected_timestep:
                raise AssertionError(assertion_error_msg)

            file_timestamp_bits_prev = file_timestamp_bits

    def _get_timestamp_regex_from_filename(self, filename):
        timestamp_bits = self._get_timestamp_bits_from_filename(filename)
        timestamp_str = timestamp_bits["timestamp_str"]
        return filename.replace(
            timestamp_str,
            self.timestamp_definitions[len(timestamp_str)]["generic_regexp"],
        )

    def _get_expected_timestep_from_filename(self, filename):
        timestamp_bits = self._get_timestamp_bits_from_filename(filename)
        timestamp_str = timestamp_bits["timestamp_str"]
        return self.timestamp_definitions[len(timestamp_str)]["expected_timestep"]

    def _get_timestamp_bits_from_filename(self, filename):
        regex_matches = re.match(self.time_period_regex, filename)
        timestamp_str = regex_matches.group(1)
        self._check_time_period_valid(timestamp_str)
        start_time = regex_matches.group(2)
        end_time = regex_matches.group(3)
        timestep = self.timestamp_definitions[len(timestamp_str)]["expected_timestep"]
        return {
            "timestamp_str": timestamp_str,
            "timestart_str": start_time,
            "timeend_str": end_time,
            "expected_timestep": timestep,
        }

    def _process_load_data_from_identifiers_warnings(self, w):
        area_cella_warn = "Missing CF-netCDF measure variable 'areacella'"
        area_cello_warn = "Missing CF-netCDF measure variable 'areacello'"
        for warn in w:
            if any(
                [m in str(warn.message) for m in (area_cella_warn, area_cello_warn)]
            ):
                self._add_areacell_measure(warn, self.areacell_var)
            else:
                logger.warning(warn.message)

    def _add_areacell_measure(self, original_warn, area_variable):
        try:
            area_cube = self.get_metadata_cube(area_variable).cube
            area_measure = iris.coords.CellMeasure(
                area_cube.core_data(),
                standard_name=area_cube.standard_name,
                long_name=area_cube.long_name,
                var_name=area_cube.var_name,
                units=area_cube.units,
                attributes=area_cube.attributes,
                measure="area",
            )
            self.cube.add_cell_measure(
                area_measure, data_dims=[self.lat_dim_number, self.lon_dim_number]
            )
        except Exception as e:  # pylint:disable=broad-except
            error_msg = str(
                original_warn.message
            ) + ". Tried to add {} cube but another exception was raised: {}".format(
                area_variable, str(e)
            )
            logger.debug(error_msg)

    def get_metadata_cube(self, metadata_variable, cube=None):
        """
        Load a metadata cube from self's attributes.

        Parameters
        ----------
        metadata_variable : str
            the name of the metadata variable to get, as it appears in the filename.

        cube : :obj:`SCMCube`
            Optionally, pass in an already loaded metadata cube to link it to
            currently loaded cube.

        Returns
        -------
        type(self)
            instance of self which has been loaded from the file containing the
            metadata variable of interest.

        Raises
        ------
        TypeError
            ``cube`` is not an :obj:`ScmCube`
        """
        if cube is not None:
            if not isinstance(cube, SCMCube):
                raise TypeError("cube must be an SCMCube instance")

            self._metadata_cubes[metadata_variable] = cube

        return self._metadata_cubes[metadata_variable]

    def get_scm_timeseries(
        self,
        surface_fraction_cube=None,
        areacell_scmcube=None,
        regions=None,
        lazy=False,
    ):
        """
        Get SCM relevant timeseries from ``self``.

        Parameters
        ----------
        surface_fraction_cube : :obj:`SCMCube`, optional
            surface fraction data which is defines surface fraction weights. If
            ``None``, we try to load the land surface fraction automatically.

        areacell_scmcube : :obj:`SCMCube`, optional
            cell area data which is used to take the latitude-longitude mean of the
            cube's data. If ``None``, we try to load this data automatically and if
            that fails we fall back onto ``iris.analysis.cartography.area_weights``.

        regions : list[str]
            List of regions to use. If ``None`` then
            ``netcdf_scm.regions.DEFAULT_REGIONS`` is used.

        lazy : bool
            Should I process the data lazily? This can be slow as data has to be read
            off disk multiple time.

        Returns
        -------
        :obj:`openscm.io.ScmDataFrame`
            An OpenSCM DataFrame instance with the data in the ``data`` attribute and
            metadata in the ``metadata`` attribute.
        """
        regions = regions if regions is not None else DEFAULT_REGIONS
        scm_timeseries_cubes = self.get_scm_timeseries_cubes(
            surface_fraction_cube=surface_fraction_cube,
            areacell_scmcube=areacell_scmcube,
            regions=regions,
            lazy=lazy,
        )

        return self.convert_scm_timeseries_cubes_to_openscmdata(scm_timeseries_cubes)

    def get_scm_timeseries_weights(
        self, surface_fraction_cube=None, areacell_scmcube=None, regions=None
    ):
        """
        Get the scm timeseries weights

        Parameters
        ----------
        surface_fraction_cube : :obj:`SCMCube`, optional
            land surface fraction data which is used to determine whether a given
            gridbox is land or ocean. If ``None``, we try to load the land surface fraction automatically.

        areacell_scmcube : :obj:`SCMCube`, optional
            cell area data which is used to take the latitude-longitude mean of the
            cube's data. If ``None``, we try to load this data automatically and if
            that fails we fall back onto ``iris.analysis.cartography.area_weights``.

        regions : list[str]
            List of regions to use. If ``None`` then
            ``netcdf_scm.regions.DEFAULT_REGIONS`` is used.

        Returns
        -------
        dict
            Dictionary of region name-weights key-value pairs
        """
        if areacell_scmcube is not None:
            # set area cube to appropriate variable
            self.get_metadata_cube(self.areacell_var, cube=areacell_scmcube)

        if surface_fraction_cube is not None:
            # set area cube to appropriate variable
            self.get_metadata_cube(
                self.surface_fraction_var, cube=surface_fraction_cube
            )

        if self._weight_calculator is None:
            self._weight_calculator = CubeWeightCalculator(self)

        regions = regions if regions is not None else DEFAULT_REGIONS
        scm_weights = self._weight_calculator.get_weights(regions)

        return scm_weights

    def get_scm_timeseries_cubes(
        self,
        surface_fraction_cube=None,
        areacell_scmcube=None,
        regions=None,
        lazy=False,
    ):
        """
        Get SCM relevant cubes

        The effective areas used for each of the regions are added as auxillary
        co-ordinates of each timeseries cube.

        If global, Northern Hemisphere and Southern Hemisphere land cubes are
        calculated, then three auxillary co-ordinates are also added to each cube:
        ``land_fraction``, ``land_fraction_northern_hemisphere`` and
        ``land_fraction_southern_hemisphere``. These co-ordinates document the area
        fraction that was considered to be land when the cubes were crunched i.e.
        ``land_fraction`` is the fraction of the entire globe which was considered to
        be land, ``land_fraction_northern_hemisphere`` is the fraction of the Northern
        Hemisphere which was considered to be land and
        ``land_fraction_southern_hemisphere`` is the fraction of the Southern
        Hemisphere which was considered to be land.

        Parameters
        ----------
        surface_fraction_cube : :obj:`SCMCube`, optional
            land surface fraction data which is used to determine whether a given
            gridbox is land or ocean. If ``None``, we try to load the land surface fraction automatically.

        areacell_scmcube : :obj:`SCMCube`, optional
            cell area data which is used to take the latitude-longitude mean of the
            cube's data. If ``None``, we try to load this data automatically and if
            that fails we fall back onto ``iris.analysis.cartography.area_weights``.

        regions : list[str]
            List of regions to use. If ``None`` then
            ``netcdf_scm.regions.DEFAULT_REGIONS`` is used.

        lazy : bool
            Should I process the data lazily? This can be slow as data has to be read
            off disk multiple time.

        Returns
        -------
        dict
            Cubes, with latitude-longitude mean data as appropriate for each of the
            SCM relevant regions.
        """
        regions = regions if regions is not None else DEFAULT_REGIONS
        scm_timeseries_weights = self.get_scm_timeseries_weights(
            surface_fraction_cube=surface_fraction_cube,
            areacell_scmcube=areacell_scmcube,
            regions=regions,
        )

        def crunch_timeseries(region, weights):
            scm_cube = take_lat_lon_mean(self, weights)
            scm_cube = self._add_metadata_to_region_timeseries_cube(scm_cube, region)

            ws = weights.shape
            if ws == self.lat_lon_shape or ws[::-1] == self.lat_lon_shape:
                weights_area_slice = weights
            else:
                area_slicer = [
                    slice(None)
                    if i in [self.lat_dim_number, self.lon_dim_number]
                    else 0
                    for i in range(len(weights.shape))
                ]
                weights_area_slice = weights[tuple(area_slicer)]
                wass = weights_area_slice.shape
                confused_shape = (
                    wass != self.lat_lon_shape and wass[::-1] != self.lat_lon_shape
                )
                if confused_shape:  # pragma: no cover
                    raise AssertionError("Can't work out area shapes")

            area = np.sum(weights_area_slice)
            if region == "World" and self.netcdf_scm_realm in ("land", "ocean"):
                # yuck hard-coding to be removed when addressing
                # https://gitlab.com/znicholls/netcdf-scm/issues/103
                area *= 1 / 100  # correct for sftlf weights being 0-100
            elif "Land" in region:
                area *= 1 / 100  # correct for sftlf weights being 0-100
            elif "Ocean" in region or "El Nino N3.4" in region:
                # El Nino N3.4 hard-coding to be removed when addressing
                # https://gitlab.com/znicholls/netcdf-scm/issues/103
                area *= 1 / 100  # correct for sftlf weights being 0-100

            return region, scm_cube, area

        memory_error = False
        if not lazy:
            try:
                crunch_list = self._crunch_in_memory(
                    crunch_timeseries, scm_timeseries_weights
                )
            except MemoryError:
                logger.warning(
                    "Data won't fit in memory, will process lazily (hence slowly)"
                )
                memory_error = True

        if lazy or memory_error:
            if lazy:
                logger.info("Forcing lazy crunching")

            data_dir = dirname(self.info["files"][0])
            self.__init__()
            self.load_data_in_directory(data_dir)
            scm_timeseries_weights = self.get_scm_timeseries_weights(
                surface_fraction_cube=surface_fraction_cube,
                areacell_scmcube=areacell_scmcube,
                regions=regions,
            )
            crunch_list = self._crunch_serial(crunch_timeseries, scm_timeseries_weights)

        timeseries_cubes = {region: ts_cube for region, ts_cube, _ in crunch_list}
        areas = {region: area for region, _, area in crunch_list if area is not None}
        timeseries_cubes = self._add_areas(timeseries_cubes, areas)
        timeseries_cubes = self._add_land_fraction(timeseries_cubes, areas)
        return timeseries_cubes

    def _crunch_in_memory(self, crunch_timeseries, scm_regions):
        # calculating lat-lon mean in parallel could go here
        self._ensure_data_realised()
        logger.debug("Crunching SCM timeseries in memory")
        return self._crunch_serial(crunch_timeseries, scm_regions)

    @staticmethod
    def _crunch_serial(crunch_timeseries, scm_regions):
        return [
            crunch_timeseries(region, weights)
            for region, weights in scm_regions.items()
        ]

    def _ensure_data_realised(self):
        # force the data to realise
        if self.cube.has_lazy_data():
            self.cube.data  # pylint:disable=pointless-statement

    def _add_areas(self, timeseries_cubes, areas):
        for cube in timeseries_cubes.values():
            for region, area in areas.items():
                area_key = self._convert_region_to_area_key(region)
                cube.cube.add_aux_coord(
                    iris.coords.AuxCoord(
                        area, long_name=area_key, units=self._area_weights_units
                    )
                )

        return timeseries_cubes

    @staticmethod
    def _convert_region_to_area_key(region):
        return "area_{}".format(region.lower().replace("|", "_").replace(" ", "_"))

    @staticmethod
    def _add_land_fraction(timeseries_cubes, areas):
        add_land_frac = all([r in areas for r in _LAND_FRACTION_REGIONS])

        if add_land_frac:
            for cube in timeseries_cubes.values():
                extensions = {
                    "land_fraction": "",
                    "land_fraction_northern_hemisphere": "|Northern Hemisphere",
                    "land_fraction_southern_hemisphere": "|Southern Hemisphere",
                }
                fractions = {}
                for k, ext in extensions.items():
                    closed_sum = np.isclose(
                        areas["World{}".format(ext)],
                        areas["World{}|Land".format(ext)]
                        + areas["World{}|Ocean".format(ext)],
                        rtol=1e-3,
                    )
                    if not closed_sum:  # pragma: no cover
                        raise AssertionError(
                            "Ocean and land area sums don't equal total..."
                        )

                    fractions[k] = (
                        areas["World{}|Land".format(ext)] / areas["World{}".format(ext)]
                    )
                fractions = {
                    k: areas["World{}|Land".format(ext)] / areas["World{}".format(ext)]
                    for k, ext in extensions.items()
                }
                for k, v in fractions.items():
                    cube.cube.add_aux_coord(
                        iris.coords.AuxCoord(v, long_name=k, units=1)
                    )
        else:
            logger.warning(
                "Not calculating land fractions as all required cubes are not "
                "available"
            )

        return timeseries_cubes

    def _add_metadata_to_region_timeseries_cube(self, scmcube, region):
        has_root_dir = (
            hasattr(self, "root_dir")  # pylint:disable=no-member
            and self.root_dir is not None  # pylint:disable=no-member
        )
        if has_root_dir:
            source_file_info = "Files: {}".format(
                [
                    p.replace(self.root_dir, "")  # pylint:disable=no-member
                    for p in self.info["files"]
                ]
            )
        else:
            source_file_info = "Files: {}".format(
                [basename(p) for p in self.info["files"]]
            )
        if "metadata" in self.info:
            source_meta = {}
            for k, v in self.info["metadata"].items():
                if has_root_dir:
                    source_meta[k] = [
                        "{}".format(
                            p.replace(self.root_dir, "")  # pylint:disable=no-member
                        )
                        for p in v["files"]
                    ]
                else:
                    source_meta[k] = ["{}".format(basename(p)) for p in v["files"]]
            source_file_info = "; ".join(
                [source_file_info]
                + ["{}: {}".format(k, v) for k, v in source_meta.items()]
            )

        scmcube.cube.attributes[
            "crunch_netcdf_scm_version"
        ] = "{} (more info at gitlab.com/znicholls/netcdf-scm)".format(__version__)
        scmcube.cube.attributes["crunch_source_files"] = source_file_info
        scmcube.cube.attributes["region"] = region
        scmcube.cube.attributes.update(self._get_scm_timeseries_ids())

        return scmcube

    def get_area_weights(self, areacell_scmcube=None):
        """
        Get area weights for this cube

        Parameters
        ----------
        areacell_scmcube : :obj:`ScmCube`
            :obj:`ScmCube` containing areacell data. If ``None``, we calculate the
            weights using iris.

        Returns
        -------
        np.ndarray
            Weights on the cube's latitude-longitude grid.

        Raises
        ------
        iris.exceptions.CoordinateMultiDimError
            The cube's co-ordinates are multi-dimensional and we don't have cell area
            data.

        ValueError
            Area weights units are not as expected (contradict with
            ``self._area_weights_units``).
        """
        areacell_scmcube = self._get_areacell_scmcube(areacell_scmcube)

        if areacell_scmcube is not None:
            areacell_units = areacell_scmcube.cube.units
            if self._area_weights_units != areacell_units:
                raise ValueError(
                    "Your weights need to be in {} but your areacell cube has units of {}".format(
                        self._area_weights_units, areacell_units
                    )
                )
            area_weights = areacell_scmcube.cube.data
            if cube_lat_lon_grid_compatible_with_array(self, area_weights):
                return area_weights
            logger.exception("Area weights incompatible with lat lon grid")

        logger.warning(
            "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
        )
        if self._area_weights_units != "m**2":
            raise ValueError(
                "iris.analysis.cartography only returns weights in m**2 but your weights need to be {}".format(
                    self._area_weights_units
                )
            )
        try:
            lat_lon_slice = next(self.cube.slices([self.lat_name, self.lon_name]))
            iris_weights = iris.analysis.cartography.area_weights(lat_lon_slice)
        except ValueError:
            logger.warning("Guessing latitude and longitude bounds")
            try:
                self.cube.coord("latitude").guess_bounds()
                self.cube.coord("longitude").guess_bounds()
            except CoordinateMultiDimError:
                error_msg = (
                    "iris does not yet support multi-dimensional co-ordinates, you "
                    "will need your data's cell area information before you can crunch"
                )
                raise CoordinateMultiDimError(error_msg)
            lat_lon_slice = next(self.cube.slices([self.lat_name, self.lon_name]))
            iris_weights = iris.analysis.cartography.area_weights(lat_lon_slice)

        return iris_weights

    def _get_areacell_scmcube(self, areacell_scmcube):
        try:
            areacell_scmcube = self.get_metadata_cube(
                self.areacell_var, cube=areacell_scmcube
            )
            if not isinstance(areacell_scmcube.cube, iris.cube.Cube):
                logger.warning(
                    "areacell cube which was found has cube attribute which isn't an iris cube"
                )
            else:
                return areacell_scmcube
        except (
            iris.exceptions.ConstraintMismatchError,
            AttributeError,
            TypeError,
            OSError,
            NotImplementedError,
            KeyError,
        ) as e:
            logger.debug("Could not calculate areacell, error message: %s", e)

        return None

    def convert_scm_timeseries_cubes_to_openscmdata(
        self, scm_timeseries_cubes, out_calendar=None
    ):
        """
        Convert dictionary of SCM timeseries cubes to an :obj:`ScmDataFrame`

        Parameters
        ----------
        scm_timeseries_cubes : dict
            Dictionary of "region name"-:obj:`ScmCube` key-value pairs.

        out_calendar : str
            Calendar to use for the time axis of the output :obj:`ScmDataFrame`

        Returns
        -------
        :obj:`ScmDataFrame`
            :obj:`ScmDataFrame` containing the data from the SCM timeseries cubes

        Raises
        ------
        NotImplementedError
            The (original) input data has dimensions other than time, latitude and
            longitude (so the data to convert has dimensions other than time).
        """
        data = []
        columns = {mc: [] for mc in _SCM_TIMESERIES_META_COLUMNS}
        for scm_cube in scm_timeseries_cubes.values():
            try:
                data.append(get_cube_timeseries_data(scm_cube, realise_data=True))
            except AssertionError:
                # blocked until we work out how to handle extra coord information in
                # SCMDataFrame
                raise NotImplementedError(
                    "Cannot yet get SCM timeseries for data with dimensions other "
                    "than time, latitude and longitude"
                )
            for column_name, column_values in columns.items():
                column_values.append(scm_cube.cube.attributes[column_name])

        data = np.vstack(data).T

        time_index, out_calendar = self._get_openscmdata_time_axis_and_calendar(
            scm_timeseries_cubes, out_calendar=out_calendar
        )
        unit = str(self.cube.units).replace("-", "^-")
        if unit == "1":
            unit = "dimensionless"  # ensure units behave with pint

        output = ScmDataFrame(
            data,
            index=time_index,
            columns={**{"unit": str(unit), "model": "unspecified"}, **columns},
        )
        try:
            output.metadata["calendar"] = out_calendar
        except AttributeError:
            output.metadata = {"calendar": out_calendar}

        return self._add_metadata_to_scmdataframe(output, scm_timeseries_cubes)

    @staticmethod
    def _add_metadata_to_scmdataframe(scmdf, scm_timeseries_cubes):
        for i, scm_cube in enumerate(scm_timeseries_cubes.values()):
            for coord in scm_cube.cube.coords():
                if coord.standard_name in ["time", "latitude", "longitude", "height"]:
                    continue

                if coord.long_name.startswith("land_fraction"):
                    new_col = coord.long_name
                    new_val = coord.points.squeeze()
                else:  # pragma: no cover
                    # this is really how it should work for land_fraction too but we don't
                    # have a stable solution for parameter handling in OpenSCMDataFrame yet so
                    # I've done the above instead
                    new_col = "{} ({})".format(coord.long_name, str(coord.units))
                    new_val = float(coord.points.squeeze())

                if not i:
                    scmdf.metadata[new_col] = new_val
                elif scmdf.metadata[new_col] != new_val:  # pragma: no cover
                    raise AssertionError("Cubes have different metadata...")

            for k, v in scm_cube.cube.attributes.items():
                if k in scmdf.meta:
                    continue
                if not i:
                    scmdf.metadata[k] = v
                elif scmdf.metadata[k] != v:  # pragma: no cover
                    raise AssertionError("Cubes have different metadata...")

        return scmdf

    def _get_scm_timeseries_ids(self):
        output = {}
        for k in _SCM_TIMESERIES_META_COLUMNS:
            if k == "region":
                continue  # handled elsewhere
            if k == "variable_standard_name":
                output[k] = self.cube.standard_name
                continue
            attr_to_check = self._scm_timeseries_id_map[k]
            val = getattr(self, attr_to_check) if hasattr(self, attr_to_check) else None
            if k == "variable":
                if val is not None:
                    output[k] = val
                else:
                    warn_msg = (
                        "Could not determine {}, filling with "
                        "standard_name".format(k)
                    )
                    logger.warning(warn_msg)
                    output[k] = self.cube.standard_name
                continue
            if val is not None:
                output[k] = val
            else:
                warn_msg = "Could not determine {}, filling with 'unspecified'".format(
                    k
                )
                logger.warning(warn_msg)
                output[k] = "unspecified"

        return output

    def _get_openscmdata_time_axis_and_calendar(
        self, scm_timeseries_cubes, out_calendar
    ):
        if out_calendar is None:
            out_calendar = self.cube.coords("time")[0].units.calendar

        time_axes = [
            get_scm_cube_time_axis_in_calendar(scm_cube, out_calendar)
            for scm_cube in scm_timeseries_cubes.values()
        ]
        assert_all_time_axes_same(time_axes)
        time_axis = time_axes[0]
        if isinstance(time_axis[0], cftime.datetime):
            # inspired by xarray, should make a PR back in there...
            if out_calendar not in {"standard", "gregorian", "proleptic_gregorian"}:
                logger.warning(
                    "Performing lazy conversion to datetime for calendar: %s. This "
                    "may cause subtle errors in operations that depend on the length "
                    "of time between dates",
                    out_calendar,
                )
            time_axis = _vector_cftime_conversion(time_axis)
        else:
            pass  # leave openscm to handle from here

        # As we sometimes have to deal with long timeseries, we force the index to be
        # pd.Index and not pd.DatetimeIndex. We can't use DatetimeIndex because of a
        # pandas limitation, see
        # http://pandas-docs.github.io/pandas-docs-travis/timeseries.html#timestamp-limitations
        return pd.Index(time_axis, dtype="object", name="time"), out_calendar

    def _check_time_period_valid(self, time_period_str):
        """
        Check that a time_period identifier string is valid.

        Parameters
        ----------
        time_period_str : str
            Time period string to check.

        Raises
        ------
        ValueError
            If the time period is not in a valid format.
        """
        if "_" in time_period_str:
            self._raise_time_period_invalid_error(time_period_str)

        if "-" in time_period_str:
            dates = time_period_str.split("-")
            if (len(dates) != 2) or (len(dates[0]) != len(dates[1])):
                self._raise_time_period_invalid_error(time_period_str)
        else:
            dates = [time_period_str]

        try:
            time_format = self.timestamp_definitions[len(dates[0])]["datetime_str"]
        except KeyError:
            self._raise_time_period_invalid_error(time_period_str)
        for date in dates:
            try:
                datetime.strptime(date, time_format)
            except ValueError:
                self._raise_time_period_invalid_error(time_period_str)

        if len(dates) == 2:
            start_date = datetime.strptime(dates[0], time_format)
            end_date = datetime.strptime(dates[1], time_format)
            if start_date >= end_date:
                self._raise_time_period_invalid_error(time_period_str)

    @staticmethod
    def _raise_time_period_invalid_error(time_period_str):
        message = "Your time_period indicator ({}) does not look right".format(
            time_period_str
        )
        raise ValueError(message)


class _CMIPCube(SCMCube, ABC):
    """Base class for cubes which follow a CMIP data reference syntax"""

    time_period = None
    """
    str: The timespan of the data stored by this cube

    The string follows the cube's data reference syntax.
    """

    filename_bits_separator = "_"
    """
    str: Character used to separate different parts of the filename

    This character is used to separate different metadata in the filename. For
    example, if ``filename_bits_separator`` is an underscore, then the filename could
    have a component like "bcc_1990-2010" and the underscore would separate "bcc" (the
    modelling centre which produced the file) from "1990-2010" (the time period of the
    file).
    """

    def load_data_from_path(self, filepath, process_warnings=True):
        """
        Load data from a path.

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.

        process_warnings : bool
            Should I process warnings to add e.g. missing metadata information?
        """
        load_data_from_identifiers_args = self.get_load_data_from_identifiers_args_from_filepath(
            filepath
        )
        self.load_data_from_identifiers(
            process_warnings=process_warnings, **load_data_from_identifiers_args
        )

    def _load_and_concatenate_files_in_directory(
        self, directory, process_warnings=True
    ):
        super()._load_and_concatenate_files_in_directory(
            directory, process_warnings=process_warnings
        )
        self._add_time_period_from_files_in_directory(directory)

    def _add_time_period_from_files_in_directory(self, directory):
        self._check_data_names_in_same_directory(directory)

        loaded_files = sorted(os.listdir(directory))
        strt = self._get_timestamp_bits_from_filename(loaded_files[0])["timestart_str"]
        end = self._get_timestamp_bits_from_filename(loaded_files[-1])["timeend_str"]
        self._time_id = self.time_period_separator.join([strt, end])

    def get_load_data_from_identifiers_args_from_filepath(self, filepath):
        """
        Get the set of identifiers to use to load data from a filepath.

        Parameters
        ----------
        filepath : str
            The filepath from which to load the data.

        Returns
        -------
        dict
            Set of arguments which can be passed to
            ``self.load_data_from_identifiers`` to load the data in the filepath.

        Raises
        ------
        ValueError
            Path and filename contradict each other
        """
        path_ids = self.process_path(dirname(filepath))
        name_ids = self.process_filename(basename(filepath))
        for key, value in path_ids.items():
            if (key in name_ids) and (value != name_ids[key]):
                error_msg = (
                    "Path and filename do not agree:\n"
                    "    - path {0}: {1}\n"
                    "    - filename {0}: {2}\n".format(key, value, name_ids[key])
                )
                raise ValueError(error_msg)

        return {**path_ids, **name_ids}

    @abstractmethod
    def process_path(self, path):
        """
        Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """

    @abstractmethod
    def process_filename(self, filename):
        """
        Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """

    def load_data_from_identifiers(self, process_warnings=True, **kwargs):
        """
        Load data using key identifiers.

        The identifiers are used to determine the path of the file to load. The file
        is then loaded into an iris cube which can be accessed through ``self.cube``.

        Parameters
        ----------
        process_warnings : bool
            Should I process warnings to add e.g. missing metadata information?

        kwargs : any
            Arguments which can then be processed by
            ``self.get_filepath_from_load_data_from_identifiers_args`` and
            ``self.get_variable_constraint`` to determine the full
            filepath of the file to load and the variable constraint to use.
        """
        with warnings.catch_warnings(record=True) as w:
            # iris v2.2.0 under py3.7 raises a DeprecationWarning about using collections, see https://github.com/SciTools/iris/pull/3320
            warnings.filterwarnings(
                "ignore", category=DeprecationWarning, module=r".*collections.*"
            )
            fpath = self.get_filepath_from_load_data_from_identifiers_args(**kwargs)
            self._load_cube(fpath, self.get_variable_constraint())

        if w and process_warnings:
            self._process_load_data_from_identifiers_warnings(w)

    def get_metadata_cube(self, metadata_variable, cube=None):
        """
        Load a metadata cube from self's attributes.

        Parameters
        ----------
        metadata_variable : str
            the name of the metadata variable to get, as it appears in the filename.

        cube : :obj:`SCMCube`
            Optionally, pass in an already loaded metadata cube to link it to
            currently loaded cube

        Returns
        -------
        type(self)
            instance of self which has been loaded from the file containing the
            metadata variable of interest.

        Raises
        ------
        TypeError
            ``cube`` is not an :obj:`ScmCube`
        """
        if cube is not None:
            return super().get_metadata_cube(metadata_variable, cube=cube)

        if metadata_variable not in self._metadata_cubes:
            load_args = self._get_metadata_load_arguments(metadata_variable)

            cube = type(self)()
            cube._metadata_cubes = {  # pylint:disable=protected-access
                k: v for k, v in self._metadata_cubes.items() if k != metadata_variable
            }
            cube.load_data_from_identifiers(**load_args)

            return super().get_metadata_cube(metadata_variable, cube=cube)
        return super().get_metadata_cube(metadata_variable)

    @abstractmethod
    def get_filepath_from_load_data_from_identifiers_args(self, **kwargs):
        """
        Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        This function should, in most cases, call ``self.get_data_directory`` and
        ``self.get_data_filename``.

        Parameters
        ----------
        **kwargs
            Arguments, initially passed to ``self.load_data_from_identifiers`` from which the full
            filepath of the file to load should be determined.

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.
        """

    def get_variable_constraint(self):
        """
        Get the iris variable constraint to use when loading data with ``self.load_data_from_identifiers``.

        Returns
        -------
        :obj:`iris.Constraint`
            constraint to use which ensures that only the variable of interest is
            loaded.
        """
        # thank you Duncan!!
        # https://github.com/SciTools/iris/issues/2107#issuecomment-246644471
        return iris.Constraint(
            cube_func=(
                lambda c: c.var_name
                == np.str(self._variable_name_for_constraint_loading)
            )
        )

    @classmethod
    def get_data_reference_syntax(cls, **kwargs):
        """
        Get data reference syntax for this cube

        Parameters
        ----------
        kwargs : str
            Attributes of the cube to set before generating the example data reference
            syntax.

        Returns
        -------
        str
            Example of the full path to a file for the given ``kwargs`` with this
            cube's data reference syntax.
        """
        helper = cls()
        for a in dir(helper):
            try:
                if callable(a) or callable(getattr(helper, a)):
                    continue
                if a == "filename_bits_separator" or a.startswith("_"):
                    continue
                new_separator = "-" if cls.filename_bits_separator == "_" else "_"
                setattr(
                    helper, a, a.replace(cls.filename_bits_separator, new_separator)
                )
            except AttributeError:
                continue

        for k, v in kwargs.items():
            setattr(helper, k, v)

        return join(helper.get_data_directory(), helper.get_data_filename())

    def get_data_directory(self):
        """
        Get the path to a data file from self's attributes.

        This can take multiple forms, it may just return a previously set
        filepath attribute or it could combine a number of different metadata
        elements (e.g. model name, experiment name) to create the data path.

        Returns
        -------
        str
            path to the data file from which this cube has been/will be loaded

        Raises
        ------
        OSError
            The data directory cannot be determined
        """
        try:
            return self._get_data_directory()
        except TypeError:  # some required attributes still None
            raise OSError("Could not determine data directory")

    @abstractmethod
    def _get_data_directory(self):
        pass  # pragma: no cover

    def get_data_filename(self):
        """
        Get the name of a data file from self's attributes.

        This can take multiple forms, it may just return a previously set
        filename attribute or it could combine a number of different metadata
        elements (e.g. model name, experiment name) to create the data name.

        Returns
        -------
        str
            name of the data file from which this cube has been/will be loaded.

        Raises
        ------
        OSError
            The data directory cannot be determined
        """
        try:
            return self._get_data_filename()
        except TypeError:  # some required attributes still None
            raise OSError("Could not determine data filename")

    @abstractmethod
    def _get_data_filename(self):
        pass  # pragma: no cover

    @abstractmethod
    def _get_metadata_load_arguments(self, metadata_variable):
        """
        Get the arguments to load a metadata file from self's attributes.

        This can take multiple forms, it may just return a previously set
        metada_filename attribute or it could combine a number of different
        metadata elements (e.g. model name, experiment name) to create the
        metadata filename.

        Parameters
        ----------
        metadata_variable : str
            the name of the metadata variable to get, as it appears in the filename.

        Returns
        -------
        dict
            dictionary containing all the arguments to pass to ``self.load_data_from_identifiers``
            required to load the desired metadata cube.
        """
        raise NotImplementedError()

    @staticmethod
    def _raise_path_error(path):
        raise ValueError("Path does not look right: {}".format(path))

    @staticmethod
    def _raise_filename_error(filename):
        raise ValueError("Filename does not look right: {}".format(filename))

    @abstractproperty
    def _variable_name_for_constraint_loading(self):
        """Variable name to use when loading an iris cube with a constraint"""

    @abstractproperty
    def _time_id(self):
        """Accessor for getting the time id (whose name varies with drs)"""

    @_time_id.setter
    def _time_id(self, value):
        """Accessor for setting the time id (whose name varies with drs)"""
        raise NotImplementedError  # pragma: no cover


class MarbleCMIP5Cube(_CMIPCube):
    """
    Subclass of ``SCMCube`` which can be used with the ``cmip5`` directory on marble.

    This directory structure is very similar, but not quite identical, to the
    recommended CMIP5 directory structure described in section 3.1 of the
    `CMIP5 Data Reference Syntax <https://cmip.llnl.gov/cmip5/docs/cmip5_data_reference_syntax_v1-00_clean.pdf>`_.
    """

    root_dir = None
    """
    str: The root directory of the database i.e. where the cube should start its path

    e.g. ``/home/users/usertim/cmip5_25x25``
    """

    activity = None
    """str: The activity for which we want to load data e.g. 'cmip5'"""

    experiment = None
    """str: The experiment for which we want to load data e.g. '1pctCO2'"""

    mip_table = None
    """str: The mip_table for which we want to load data e.g. 'Amon'"""

    variable_name = None
    """str: The variable for which we want to load data e.g. 'tas'"""

    model = None
    """str: The model for which we want to load data e.g. 'CanESM2'"""

    ensemble_member = None
    """str: The ensemble member for which we want to load data e.g. 'r1i1p1'"""

    time_period = None
    """
    str: The time period for which we want to load data

    If ``None``, this information isn't included in the filename which is useful for
    loading metadata files which don't have a relevant time period.
    """

    file_ext = None
    """str: The file extension of the data file we want to load e.g. '.nc'"""

    mip_era = "CMIP5"
    """str: The MIP era to which this cube belongs"""

    _realm_key = "modeling_realm"

    @property
    def _table_name_for_metadata_vars(self):
        return "fx"

    def process_filename(self, filename):
        """
        Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split(self.filename_bits_separator)
        if len(filename_bits) == 6:
            time_period, file_ext = splitext(filename_bits[-1])
            ensemble_member = filename_bits[-2]
        elif len(filename_bits) == 5:
            time_period = None
            ensemble_member, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_name": filename_bits[0],
            "mip_table": filename_bits[1],
            "model": filename_bits[2],
            "experiment": filename_bits[3],
            "ensemble_member": ensemble_member,
            "time_period": time_period,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """
        Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 6) or any(
            [self.filename_bits_separator in d for d in dirpath_bits[-6:]]
        ):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-6])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "activity": dirpath_bits[-6],
            "variable_name": dirpath_bits[-3],
            "mip_table": dirpath_bits[-4],
            "model": dirpath_bits[-2],
            "experiment": dirpath_bits[-5],
            "ensemble_member": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(self, **kwargs):
        """
        Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the identifiers are given in Section 2 of the
        `CMIP5 Data Reference Syntax <https://cmip.llnl.gov/cmip5/docs/cmip5_data_reference_syntax_v1-00_clean.pdf>`_.

        Parameters
        ----------
        kwargs : str
            Identifiers to use to load the data

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.

        Raises
        ------
        AttributeError
            An input argument does not match with the cube's data reference syntax
        """
        for name, value in kwargs.items():
            setattr(self, name, value)

        return join(self.get_data_directory(), self.get_data_filename())

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.activity,
            self.experiment,
            self.mip_table,
            self.variable_name,
            self.model,
            self.ensemble_member,
        )

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_name,
            self.mip_table,
            self.model,
            self.experiment,
            self.ensemble_member,
        ]
        if self.time_period is not None:
            bits_to_join.append(self.time_period)

        return self.filename_bits_separator.join(bits_to_join) + self.file_ext

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "activity": self.activity,
            "experiment": self.experiment,
            "mip_table": self.table_name_for_metadata_vars,
            "variable_name": metadata_variable,
            "model": self.model,
            "ensemble_member": "r0i0p0",
            "time_period": None,
            "file_ext": self.file_ext,
        }

    @property
    def _variable_name_for_constraint_loading(self):
        return self.variable_name

    @property
    def _time_id(self):
        return self.time_period

    @_time_id.setter
    def _time_id(self, value):
        self.time_period = value


class CMIP6Input4MIPsCube(_CMIPCube):
    """
    Subclass of ``SCMCube`` which can be used with CMIP6 input4MIPs data

    The data must match the CMIP6 Forcing Datasets Summary, specifically the
    `Forcing Dataset Specifications <https://docs.google.com/document/d/1pU9IiJvPJwRvIgVaSDdJ4O0Jeorv_2ekEtted34K9cA/edit#heading=h.cn9f7982ycw6>`_.
    """

    root_dir = None
    """
    str: The root directory of the database i.e. where the cube should start its
            path

    e.g. ``/home/users/usertim/cmip6input``.
    """

    activity_id = None
    """
    str: The activity_id for which we want to load data.

    For these cubes, this will almost always be ``input4MIPs``.
    """

    mip_era = None
    """str: The mip_era for which we want to load data e.g. ``CMIP6``"""

    target_mip = None
    """str:The target_mip for which we want to load data e.g. ``ScenarioMIP``"""

    institution_id = None
    """str: The institution_id for which we want to load data e.g. ``UoM``"""

    source_id = None
    """
    str: The source_id for which we want to load data e.g.
            ``UoM-REMIND-MAGPIE-ssp585-1-2-0``

    This must include the institution_id.
    """

    realm = None
    """str: The realm for which we want to load data e.g. ``atmos``"""

    frequency = None
    """str: The frequency for which we want to load data e.g. ``yr``"""

    variable_id = None
    """str: The variable_id for which we want to load data e.g. ``mole-fraction-of-carbon-dioxide-in-air``"""

    grid_label = None
    """str: The grid_label for which we want to load data e.g. ``gr1-GMNHSH``"""

    version = None
    """str: The version for which we want to load data e.g. ``v20180427``"""

    dataset_category = None
    """str: The dataset_category for which we want to load data e.g.
            ``GHGConcentrations``"""

    time_range = None
    """
    str: The time range for which we want to load data e.g. ``2005-2100``

    If ``None``, this information isn't included in the filename which is useful
    for loading metadata files which don't have a relevant time period.
    """

    file_ext = None
    """str: The file extension of the data file we want to load e.g. ``.nc``"""

    def process_filename(self, filename):
        """
        Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split(self.filename_bits_separator)
        if len(filename_bits) == 7:
            time_range, file_ext = splitext(filename_bits[-1])
            grid_label = filename_bits[-2]
        elif len(filename_bits) == 6:
            time_range = None
            grid_label, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_id": filename_bits[0],
            "activity_id": filename_bits[1],
            "dataset_category": filename_bits[2],
            "target_mip": filename_bits[3],
            "source_id": filename_bits[4],
            "grid_label": grid_label,
            "time_range": time_range,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """
        Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 10) or any(
            [self.filename_bits_separator in d for d in dirpath_bits[-10:]]
        ):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-10])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "activity_id": dirpath_bits[-10],
            "mip_era": dirpath_bits[-9],
            "target_mip": dirpath_bits[-8],
            "institution_id": dirpath_bits[-7],
            "source_id": dirpath_bits[-6],
            "realm": dirpath_bits[-5],
            "frequency": dirpath_bits[-4],
            "variable_id": dirpath_bits[-3],
            "grid_label": dirpath_bits[-2],
            "version": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(self, **kwargs):
        """
        Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the meaning of the identifiers are given in the
        `Forcing Dataset Specifications <https://docs.google.com/document/d/1pU9IiJvPJwRvIgVaSDdJ4O0Jeorv_2ekEtted34K9cA/edit#heading=h.cn9f7982ycw6>`_.

        Parameters
        ----------
        kwargs : str
            Identifiers to use to load the data

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.

        Raises
        ------
        AttributeError
            An input argument does not match with the cube's data reference syntax
        """
        for name, value in kwargs.items():
            setattr(self, name, value)

        # TODO: do time indicator/frequency checks too and make a new method for
        # checks so can be reused by different methods
        self._check_self_consistency()

        return join(self.get_data_directory(), self.get_data_filename())

    def _check_self_consistency(self):
        if str(self.institution_id) not in str(self.source_id):
            raise AssertionError("source_id must contain institution_id")

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "activity_id": self.activity_id,
            "mip_era": self.mip_era,
            "target_mip": self.target_mip,
            "institution_id": self.institution_id,
            "source_id": self.source_id,
            "realm": self.realm,
            "frequency": "fx",
            "variable_id": metadata_variable,
            "grid_label": self.grid_label,
            "version": self.version,
            "dataset_category": self.dataset_category,
            "time_range": None,
            "file_ext": self.file_ext,
        }

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_id,
            self.activity_id,
            self.dataset_category,
            self.target_mip,
            self.source_id,
            self.grid_label,
        ]
        if self.time_range is not None:
            bits_to_join.append(self.time_range)

        return self.filename_bits_separator.join(bits_to_join) + self.file_ext

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.activity_id,
            self.mip_era,
            self.target_mip,
            self.institution_id,
            self.source_id,
            self.realm,
            self.frequency,
            self.variable_id,
            self.grid_label,
            self.version,
        )

    @property
    def _variable_name_for_constraint_loading(self):
        return self.variable_id.replace("-", "_")

    @property
    def _time_id(self):
        return self.time_range

    @_time_id.setter
    def _time_id(self, value):
        self.time_range = value


class CMIP6OutputCube(_CMIPCube):
    """
    Subclass of ``SCMCube`` which can be used with CMIP6 model output data

    The data must match the CMIP6 data reference syntax as specified in the 'File name
    template' and 'Directory structure template' sections of the
    `CMIP6 Data Reference Syntax <https://goo.gl/v1drZl>`_.
    """

    root_dir = None
    """
    str: The root directory of the database i.e. where the cube should start its
            path

    e.g. ``/home/users/usertim/cmip6_data``.
    """

    activity_id = None
    """
    str: The activity_id for which we want to load data.

    In CMIP6, this denotes the responsible MIP e.g. ``DCPP``.
    """

    mip_era = None
    """str: The mip_era for which we want to load data e.g. ``CMIP6``"""

    institution_id = None
    """str: The institution_id for which we want to load data e.g. ``CNRM-CERFACS``"""

    source_id = None
    """
    str: The source_id for which we want to load data e.g. ``CNRM-CM6-1``

    This was known as model in CMIP5.
    """

    experiment_id = None
    """str: The experiment_id for which we want to load data e.g. ``dcppA-hindcast``"""

    member_id = None
    """str: The member_id for which we want to load data e.g. ``s1960-r2i1p1f3``"""

    table_id = None
    """str: The table_id for which we want to load data. e.g. ``day``"""

    variable_id = None
    """str: The variable_id for which we want to load data e.g. ``pr``"""

    grid_label = None
    """str: The grid_label for which we want to load data e.g. ``grn``"""

    version = None
    """str: The version for which we want to load data e.g. ``v20160215``"""

    time_range = None
    """
    str: The time range for which we want to load data e.g. ``198001-198412``

    If ``None``, this information isn't included in the filename which is useful
    for loading metadata files which don't have a relevant time period.
    """

    file_ext = None
    """str: The file extension of the data file we want to load e.g. ``.nc``"""

    _scm_timeseries_id_map = {
        "variable": "variable_id",
        "climate_model": "source_id",
        "scenario": "experiment_id",
        "activity_id": "activity_id",
        "member_id": "member_id",
        "mip_era": "mip_era",
    }

    def process_filename(self, filename):
        """
        Cut a filename into its identifiers

        Parameters
        ----------
        filename : str
            The filename to process. Filename here means just the filename, no path
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input filename
        """
        filename_bits = filename.split(self.filename_bits_separator)
        if len(filename_bits) == 7:
            time_range, file_ext = splitext(filename_bits[-1])
            grid_label = filename_bits[-2]
        elif len(filename_bits) == 6:
            time_range = None
            grid_label, file_ext = splitext(filename_bits[-1])
        else:
            self._raise_filename_error(filename)

        if not file_ext:
            self._raise_filename_error(filename)

        return {
            "variable_id": filename_bits[0],
            "table_id": filename_bits[1],
            "source_id": filename_bits[2],
            "experiment_id": filename_bits[3],
            "member_id": filename_bits[4],
            "grid_label": grid_label,
            "time_range": time_range,
            "file_ext": file_ext,
        }

    def process_path(self, path):
        """
        Cut a path into its identifiers

        Parameters
        ----------
        path : str
            The path to process. Path here means just the path, no filename
            should be included.

        Returns
        -------
        dict
            A dictionary where each key is the identifier name and each value is the value of that identifier for the input path
        """
        path = path[:-1] if path.endswith(os.sep) else path
        dirpath_bits = path.split(os.sep)
        if (len(dirpath_bits) < 10) or any(
            [self.filename_bits_separator in d for d in dirpath_bits[-10:]]
        ):
            self._raise_path_error(path)

        root_dir = os.sep.join(dirpath_bits[:-10])
        if not root_dir:
            root_dir = "."

        return {
            "root_dir": root_dir,
            "mip_era": dirpath_bits[-10],
            "activity_id": dirpath_bits[-9],
            "institution_id": dirpath_bits[-8],
            "source_id": dirpath_bits[-7],
            "experiment_id": dirpath_bits[-6],
            "member_id": dirpath_bits[-5],
            "table_id": dirpath_bits[-4],
            "variable_id": dirpath_bits[-3],
            "grid_label": dirpath_bits[-2],
            "version": dirpath_bits[-1],
        }

    def get_filepath_from_load_data_from_identifiers_args(self, **kwargs):
        """
        Get the full filepath of the data to load from the arguments passed to ``self.load_data_from_identifiers``.

        Full details about the meaning of each identifier is given in Table 1 of the
        `CMIP6 Data Reference Syntax <https://goo.gl/v1drZl>`_.

        Parameters
        ----------
        kwargs : str
            Identifiers to use to load the data

        Returns
        -------
        str
            The full filepath (path and name) of the file to load.

        Raises
        ------
        AttributeError
            An input argument does not match with the cube's data reference syntax
        """
        for name, value in kwargs.items():
            setattr(self, name, value)

        return join(self.get_data_directory(), self.get_data_filename())

    def _get_metadata_load_arguments(self, metadata_variable):
        return {
            "root_dir": self.root_dir,
            "mip_era": self.mip_era,
            "activity_id": self.activity_id,
            "institution_id": self.institution_id,
            "source_id": self.source_id,
            "experiment_id": self.experiment_id,
            "member_id": self.member_id,
            "table_id": self.table_name_for_metadata_vars,
            "variable_id": metadata_variable,
            "grid_label": self.grid_label,
            "version": self.version,
            "time_range": None,
            "file_ext": self.file_ext,
        }

    def _get_data_filename(self):
        bits_to_join = [
            self.variable_id,
            self.table_id,
            self.source_id,
            self.experiment_id,
            self.member_id,
            self.grid_label,
        ]
        if self.time_range is not None:
            bits_to_join.append(self.time_range)

        return self.filename_bits_separator.join(bits_to_join) + self.file_ext

    def _get_data_directory(self):
        return join(
            self.root_dir,
            self.mip_era,
            self.activity_id,
            self.institution_id,
            self.source_id,
            self.experiment_id,
            self.member_id,
            self.table_id,
            self.variable_id,
            self.grid_label,
            self.version,
        )

    @property
    def _variable_name_for_constraint_loading(self):
        return self.variable_id.replace("-", "_")

    @property
    def _time_id(self):
        return self.time_range

    @_time_id.setter
    def _time_id(self, value):
        self.time_range = value



"""
Module which calculates the weights to be used when taking SCM-box averages

This typically requires considering both the fraction of each cell which is of the desired type (e.g. land or ocean) and the area of each cell. The combination of these two pieces of information creates the weights for each cell which are used when taking area-weighted means.
"""
import logging
import os
from functools import lru_cache

import numpy as np

from ..utils import cube_lat_lon_grid_compatible_with_array

try:
    import iris
    from iris.analysis.cartography import wrap_lons
except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from ..errors import raise_no_iris_warning

    raise_no_iris_warning()


logger = logging.getLogger(__name__)

_DEFAULT_SFTLF_FILE = "default_land_ocean_weights.nc"
DEFAULT_REGIONS = (
    "World",
    "World|Land",
    "World|Ocean",
    "World|Northern Hemisphere",
    "World|Southern Hemisphere",
    "World|Northern Hemisphere|Land",
    "World|Southern Hemisphere|Land",
    "World|Northern Hemisphere|Ocean",
    "World|Southern Hemisphere|Ocean",
)


class InvalidWeights(Exception):
    """
    Raised when a weight cannot be calculated.

    This error usually propogates. For example, if a child weight used in the
    calculation of a parent weight fails then the parent weight should also raise an
    InvalidWeights exception (unless it can be satisfactorily handled).
    """


def subtract_weights(weights_to_subtract, subtract_from):
    """
    Subtract weights from some other number

    e.g. useful to convert e.g. from fraction of land to ocean (where
    ocean fractions are 100 - land fractions)

    Parameters
    ----------
    weights_to_subtract: str
        Name of the weights to subtract. These weights are loaded at evaluation time.

    subtract_from : float
        The number from which to subtract the values of ``weights_to_invert`` (once
        loaded)

    Returns
    -------
    :func:`WeightFunc`
        WeightFunc which subtracts the input weights from ``subtract_from``
    """

    def f(weight_calculator, cube, **kwargs):  # pylint:disable=unused-argument
        base = (
            weights_to_subtract(weight_calculator, cube, **kwargs)
            if callable(weights_to_subtract)
            else weight_calculator.get_weights_array_without_area_weighting(
                weights_to_subtract
            )
        )

        return subtract_from - base

    return f


def multiply_weights(weight_a, weight_b):
    """
    Take the product of two weights

    Parameters
    ----------
    weight_a : str or WeightFunc
        If a string is provided, the weights specified by the string are retrieved.
        Otherwise the WeightFunc is evaluated at runtime

    weight_b: str or WeightFunc
        If a string is provided, the weights specified by the string are retrieved.
        Otherwise the WeightFunc is evaluated at runtime

    Returns
    -------
    :func:`WeightFunc`
        WeightFunc which multiplies the input weights
    """

    def f(weight_calculator, cube, **kwargs):
        a = (
            weight_a(weight_calculator, cube, **kwargs)
            if callable(weight_a)
            else weight_calculator.get_weights_array_without_area_weighting(weight_a)
        )
        b = (
            weight_b(weight_calculator, cube, **kwargs)
            if callable(weight_b)
            else weight_calculator.get_weights_array_without_area_weighting(weight_b)
        )
        return a * b

    return f


@lru_cache(maxsize=1)
def get_default_sftlf_cube():
    """Load NetCDF-SCM's default (last resort) surface land fraction cube"""
    return iris.load_cube(os.path.join(os.path.dirname(__file__), _DEFAULT_SFTLF_FILE))


def _check_surface_fraction_bounds_and_shape(surface_frac_data, base_cube):
    surface_frac_data_max = surface_frac_data.max()
    if np.isclose(surface_frac_data_max, 1, atol=0.3):
        logger.warning(
            "%s data max is %s, multiplying by 100 to convert units to percent",
            base_cube.surface_fraction_var,
            surface_frac_data_max,
        )
        surface_frac_data = surface_frac_data * 100

    if not cube_lat_lon_grid_compatible_with_array(base_cube, surface_frac_data):
        raise AssertionError(
            "the {} cube data must be the same shape as the cube's "
            "longitude-latitude grid".format(base_cube.surface_fraction_var)
        )

    return surface_frac_data


def get_land_weights(  # pylint:disable=unused-argument
    weight_calculator, cube, sftlf_cube=None, **kwargs
):
    """
    Get the land weights

    The weights are always adjusted to have units of percentage. If the units are
    detected to be fraction rather than percentage, they will be automatically
    adjusted and a warning will be thrown.

    If the default sftlf cube is used, it is regridded onto ``cube``'s grid using a
    linear interpolation. We hope to use an area-weighted regridding in future but at
    the moment its performance is not good enough to be put into production (
    approximately 100x slower than the linear interpolation regridding).

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    sftlf_cube : :obj:`SCMCube`
        Cube containing the surface land-fraction data

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    np.ndarray
        Land weights

    Raises
    ------
    AssertionError
        The land weights are incompatible with the cube's lat-lon grid
    """

    def _return_and_set_cache(vals):
        weight_calculator._weights_no_area_weighting[  # pylint:disable=protected-access
            "World|Land"
        ] = vals
        return vals

    if cube.netcdf_scm_realm == "ocean":
        return _return_and_set_cache(
            np.zeros(get_binary_nh_weights(weight_calculator, cube).shape)
        )

    sftlf_data = None
    try:
        sftlf_cube = cube.get_metadata_cube(cube.surface_fraction_var, cube=sftlf_cube)
        sftlf_data = sftlf_cube.cube.data
    except (OSError, KeyError):  # TODO: fix reading so KeyError not needed
        warn_msg = (
            "Land surface fraction (sftlf) data not available, using default instead"
        )
        logger.warning(warn_msg)
        try:
            def_cube_regridded = (
                get_default_sftlf_cube()
                .copy()
                .regrid(
                    cube.cube,
                    iris.analysis.Linear(),  # AreaWeighted() in future but too slow now
                )
            )
        except ValueError:  # pragma: no cover # only required for AreaWeighted() regridding
            logger.warning("Guessing bounds to regrid default sftlf data")
            cube.lat_dim.guess_bounds()
            cube.lon_dim.guess_bounds()
            def_cube_regridded = (
                get_default_sftlf_cube()
                .copy()
                .regrid(
                    cube.cube,
                    iris.analysis.Linear(),  # AreaWeighted() in future but too slow now
                )
            )

        sftlf_data = def_cube_regridded.data

    sftlf_data = _check_surface_fraction_bounds_and_shape(sftlf_data, cube)

    return _return_and_set_cache(sftlf_data)


def get_ocean_weights(  # pylint:disable=unused-argument
    weight_calculator, cube, sftof_cube=None, **kwargs
):
    """
    Get the ocean weights

    The weights are always adjusted to have units of percentage.

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    sftof_cube : :obj:`SCMCube`
        Cube containing the surface ocean-fraction data

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    np.ndarray
        Ocean weights

    Raises
    ------
    AssertionError
        The ocean weights are incompatible with the cube's lat-lon grid
    """

    def _return_and_set_cache(vals):
        weight_calculator._weights_no_area_weighting[  # pylint:disable=protected-access
            "World|Ocean"
        ] = vals
        return vals

    if cube.netcdf_scm_realm == "land":
        return _return_and_set_cache(
            0 * np.ones(get_binary_nh_weights(weight_calculator, cube).shape)
        )

    if cube.netcdf_scm_realm == "ocean":
        try:
            sftof_cube = cube.get_metadata_cube(
                cube.surface_fraction_var, cube=sftof_cube
            )
            sftof_data = sftof_cube.cube.data
            sftof_data = _check_surface_fraction_bounds_and_shape(sftof_data, cube)

            return _return_and_set_cache(sftof_data)

        except (OSError, KeyError):  # TODO: fix reading so KeyError not needed
            pass

    return _return_and_set_cache(
        100 - weight_calculator.get_weights_array_without_area_weighting("World|Land")
    )


def get_binary_nh_weights(
    weight_calculator, cube, **kwargs
):  # pylint:disable=unused-argument
    """
    Get binary weights to only include the Northern Hemisphere

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    :obj:`np.ndarray`
        Binary northern hemisphere weights
    """
    weights_nh_lat = np.array([c >= 0 for c in cube.lat_dim.points]).astype(int)
    weights_all_lon = np.ones(cube.lon_dim.points.shape)

    if len(weights_nh_lat.shape) > 1:
        weights_nh = weights_nh_lat * weights_all_lon
    else:
        weights_nh = np.outer(weights_nh_lat, weights_all_lon)

    return weights_nh


def get_nh_weights(weight_calculator, cube, **kwargs):  # pylint:disable=unused-argument
    """
    Get weights to only include the Northern Hemisphere

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    :obj:`np.ndarray`
        Northern hemisphere weights
    """
    weights_nh = get_binary_nh_weights(weight_calculator, cube, **kwargs)

    if cube.netcdf_scm_realm == "land":
        weights_nh *= (
            weight_calculator.get_weights_array_without_area_weighting("World|Land")
            / 100
        )  # maintain normalisation to 1

    elif cube.netcdf_scm_realm == "ocean":
        weights_nh *= (
            weight_calculator.get_weights_array_without_area_weighting("World|Ocean")
            / 100
        )  # maintain normalisation to 1

    weight_calculator._weights_no_area_weighting[  # pylint:disable=protected-access
        "World|Northern Hemisphere"
    ] = weights_nh

    return weights_nh


def get_sh_weights(weight_calculator, cube, **kwargs):  # pylint:disable=unused-argument
    """
    Get weights to only include the Southern Hemisphere

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    :obj:`np.ndarray`
        Southern hemisphere weights
    """
    weights_sh = 1 - get_binary_nh_weights(weight_calculator, cube, **kwargs)

    if cube.netcdf_scm_realm == "land":
        weights_sh *= (
            weight_calculator.get_weights_array_without_area_weighting("World|Land")
            / 100
        )  # maintain normalisation to 1

    elif cube.netcdf_scm_realm == "ocean":
        weights_sh *= (
            weight_calculator.get_weights_array_without_area_weighting("World|Ocean")
            / 100
        )  # maintain normalisation to 1

    weight_calculator._weights_no_area_weighting[  # pylint:disable=protected-access
        "World|Southern Hemisphere"
    ] = weights_sh

    return weights_sh


def get_weights_for_area(lower_lat, left_lon, upper_lat, right_lon):
    """
    Weights a subset of the globe using latitudes and longitudes (in degrees East)

    Iris' standard behaviour is to include any point whose bounds overlap with
    the given ranges e.g. if the range is (0, 130) then a cell whose bounds were
    (-90, 5) would be included even if its point were -42.5.

    This can be altered with the ``ignore_bounds`` keyword argument to
    ``cube.intersection``. In this case only cells whose points lie within the
    range are included so if the range is (0, 130) then a cell whose bounds were
    (-90, 5) would be excluded if its point were -42.5.

    Here we follow the ``ignore_bounds=True`` behaviour (i.e. only include if
    the point lies within the specified range). If we want to only include the
    cell if the entire box is within a point we're going to need to tweak things.
    Given this isn't available in iris, it seems to be an unusual way to do
    intersection so we haven't implemented it.

    Circular coordinates (longitude) can cross the 0E.

    Parameters
    ----------
    lower_lat : int or float
        Lower latitude bound (degrees North)

    left_lon : int or float
        Lower longitude bound (degrees East)

    upper_lat : int or float
        Upper latitude bound (degrees North)

    right_lon : int or float
        Upper longitude bound (degrees East)

    Returns
    -------
    :func:`WeightFunc`
        WeightFunc which weights out everything except the specified area
    """

    def f(weight_calculator, cube, **kwargs):  # pylint:disable=unused-argument
        lon_dim_pts = cube.lon_dim.points
        lat_dim_pts = cube.lat_dim.points

        lat_lon_size = cube.lat_lon_shape

        if len(lat_dim_pts.shape) == 1:
            lat_dim_pts = np.broadcast_to(lat_dim_pts, lat_lon_size[::-1]).T
        if len(lon_dim_pts.shape) == 1:
            lon_dim_pts = np.broadcast_to(lon_dim_pts, lat_lon_size)

        weights_lat = ((lower_lat <= lat_dim_pts) & (lat_dim_pts <= upper_lat)).astype(
            int
        )

        lon_modulus = cube.lon_dim.units.modulus
        lon_min = np.floor(lon_dim_pts.min())
        left_lon_wrapped, right_lon_wrapped = wrap_lons(
            np.array([left_lon, right_lon]), lon_min, lon_modulus
        ).astype(int)
        if left_lon_wrapped <= right_lon_wrapped:
            weights_lon = (left_lon_wrapped <= lon_dim_pts) & (
                lon_dim_pts <= right_lon_wrapped
            )
        else:
            weights_lon = (
                (lon_min <= lon_dim_pts) & (lon_dim_pts <= right_lon_wrapped)
            ) | (
                (left_lon_wrapped <= lon_dim_pts)
                & (lon_dim_pts <= lon_min + lon_modulus)
            )

        weights_lon = weights_lon.astype(int)

        # TODO: make issue in Iris about the fact that ``cube.intersection``'s errors
        # are cryptic
        error_msg = "None of the cube's {} lie within the bounds:\nquery: ({}, {})\ncube points: {}"
        if np.equal(np.sum(weights_lon), 0):
            raise ValueError(
                error_msg.format("latitudes", lower_lat, upper_lat, cube.lat_dim.points)
            )

        if np.equal(np.sum(weights_lat), 0):
            raise ValueError(
                error_msg.format("longitudes", left_lon, right_lon, cube.lon_dim.points)
            )

        weights = weights_lon * weights_lat
        return weights

    return f


def get_world_weights(
    weight_calculator, cube, **kwargs
):  # pylint:disable=unused-argument
    """
    Get weights for the world

    Parameters
    ----------
    weight_calculator : :obj:`CubeWeightCalculator`
        Cube weight calculator from which to retrieve the weights

    cube : :obj:`SCMCube`
        Cube to create weights for

    kwargs : Any
        Ignored (required for compatibility with ``CubeWeightCalculator``)

    Returns
    -------
    :obj:`np.ndarray`
        Weights which can be used for the world mean calculation
    """
    if cube.netcdf_scm_realm == "land":
        return weight_calculator.get_weights_array_without_area_weighting("World|Land")

    if cube.netcdf_scm_realm == "ocean":
        return weight_calculator.get_weights_array_without_area_weighting("World|Ocean")

    return np.ones(
        weight_calculator.get_weights_array_without_area_weighting(
            "World|Northern Hemisphere"
        ).shape
    )


"""dict: in-built functions to calculate weights for different regions without area weighting"""
WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING = {
    "World": get_world_weights,
    "World|Northern Hemisphere": get_nh_weights,
    "World|Southern Hemisphere": get_sh_weights,
    "World|Land": get_land_weights,
    "World|Ocean": get_ocean_weights,
    "World|Northern Hemisphere|Land": multiply_weights(
        get_binary_nh_weights, "World|Land"
    ),
    "World|Southern Hemisphere|Land": multiply_weights(
        subtract_weights(get_binary_nh_weights, 1), "World|Land"
    ),
    "World|Northern Hemisphere|Ocean": multiply_weights(
        get_binary_nh_weights, "World|Ocean"
    ),
    "World|Southern Hemisphere|Ocean": multiply_weights(
        subtract_weights(get_binary_nh_weights, 1), "World|Ocean"
    ),
    "World|North Atlantic Ocean": multiply_weights(
        get_weights_for_area(0, -80, 65, 0), "World|Ocean"
    ),
    # 5N-5S, 170W-120W (i.e. 190E to 240E) see
    # https://climatedataguide.ucar.edu/climate-data/nino-sst-indices-nino-12-3-34-4-oni-and-tni
    "World|El Nino N3.4": multiply_weights(
        get_weights_for_area(-5, 190, 5, 240), "World|Ocean"
    ),
}


class CubeWeightCalculator:
    """
    Computes weights for a given cube in a somewhat efficient manner.

    Previously calculated weights are cached so each set of weights is only calculated
    once. This implementation trades off some additional memory overhead for the
    ability to generate arbitary weights.

    **Adding new weights**

    Additional weights can be added to ``netcdf_scm.weights.weights``. The values in
    ``weights`` should be WeightFunc's. A WeightFunc is a function which takes a
    ScmCube, CubeWeightCalculator and any additional keyword arguments. The function
    should return a numpy array with the same dimensionality as the ScmCube.

    These WeightFunc's can be composed together to create more complex functionality
    using e.g. ``multiply_weights``.
    """

    def __init__(self, cube, **kwargs):
        """
        Initialise

        Parameters
        ----------
        cube : :obj:`ScmCube`
            cube to generate weights for

        kwargs : dict
            Any optional arguments to be passed to the WeightFunc's during evaluation.
            Possible parameters include:

                sftlf_cube : ScmCube
        """
        self.cube = cube
        self._weights_no_area_weighting = {}
        self._weights = {}
        self._area_weights = None
        self.kwargs = kwargs

    def get_weights_array_without_area_weighting(self, weights_name):
        """
        Get a single weights array without any consideration of area weighting

        Parameters
        ----------
        weights_name : str
            Region to get weights for

        Returns
        -------
        np.ndarray
            Weights

        Raises
        ------
        InvalidWeights
            If the requested weights cannot be found or evaluated
        """
        try:
            return self._weights_no_area_weighting[weights_name]
        except KeyError:
            try:
                weights_func = WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING[weights_name]
                weights = weights_func(self, self.cube, **self.kwargs)

                self._weights_no_area_weighting[weights_name] = weights
            except KeyError:
                raise InvalidWeights("Unknown weights: {}".format(weights_name))

        return weights

    def _get_area_weights(self):
        if self._area_weights is None:
            self._area_weights = self.cube.get_area_weights()

        return self._area_weights

    def get_weights_array(self, weights_name):
        """
        Get a single weights array

        If the weights have previously been calculated the precalculated result is
        returned from the cache. Otherwise the appropriate WeightFunc is called with
        any kwargs specified in the constructor.

        Parameters
        ----------
        weights_name : str
            Region to get weights for

        Returns
        -------
        ndarray[bool]
            Weights for the region specified by ``weights_name``

        Raises
        ------
        ValueError
            If the cube has no data which matches the input weights
        """
        try:
            return self._weights[weights_name]
        except KeyError:
            weights_without_area = self.get_weights_array_without_area_weighting(
                weights_name
            )
            area_weights = self._get_area_weights()
            weights = weights_without_area * area_weights

        if np.equal(np.sum(weights), 0):
            raise ValueError(
                "All weights are zero for region: `{}`".format(weights_name)
            )

        self._weights[weights_name] = weights
        return weights

    def get_weights(self, weights_names):
        """
        Get a number of weights

        Parameters
        ----------
        weights_names: list of str
            List of weights to attempt to load/calculate.

        Returns
        -------
        dict
            Dictionary where keys are weights names and values are :obj:`np.ndarray` of
            bool

        The result only contains valid weights. Any invalid weights are dropped.
        """
        weights = {}
        for name in weights_names:
            try:
                weights_for_name = self.get_weights_array(name)
                weights[name] = weights_for_name
            except InvalidWeights as e:
                logger.warning("Failed to create %s weights: %s", name, str(e))

        return weights
